// Actions.cpp : Implementation of CActions
#include "stdafx.h"
#include "GnlPpsEvntCsmr.h"
#include "Actions.h"

/////////////////////////////////////////////////////////////////////////////
// CActions


STDMETHODIMP CActions::CreatePrc(BSTR bsCmdLine, INT iShowWindow)
{
	// purpose = execute a CreateProcess api call using the passed parameters
	HRESULT hr = S_OK;
	PROCESS_INFORMATION oPrcInfo = {NULL};
	try {
		// validate parameters
		if (!bsCmdLine || !SysStringLen(bsCmdLine) /* || !wNumObjects */) { hr = E_INVALIDARG; throw hr; }

		STARTUPINFO oStrUpInfo = {sizeof(oStrUpInfo), NULL, NULL};  // minimum required for CreateProcess call
		oStrUpInfo.wShowWindow = (WORD)iShowWindow;  // automation clients don't like WORD method parameter as a type
		BOOL br = CreateProcess(NULL, (LPTSTR)bsCmdLine, NULL, NULL, FALSE,
								0, NULL, // | CREATE_UNICODE_ENVIRONMENT, pvEnvBlk,
								NULL, &oStrUpInfo, &oPrcInfo);
		if (!br) {hr = HRESULT_FROM_WIN32(GetLastError()); throw hr;}
	}

	catch(...) {
		OutputDebugStringEx(hr, _bstr_t(_T("CreatePrc")));
		if (SUCCEEDED(hr)) throw;  // pass unexpected exceptions on to the outer block
	}

	if (oPrcInfo.hThread && (INVALID_HANDLE_VALUE != oPrcInfo.hThread)) {
		BOOL br = CloseHandle(oPrcInfo.hThread); // if (!br) throw;
		oPrcInfo.hThread = NULL;
	}
	if (oPrcInfo.hProcess && (INVALID_HANDLE_VALUE != oPrcInfo.hProcess)) {
		BOOL br = CloseHandle(oPrcInfo.hProcess); // if (!br) throw;
		oPrcInfo.hProcess = NULL;
	}

	return hr;
}


STDMETHODIMP CActions::XmlHttpSqlInsert(BSTR bsRdmsHost, BSTR bsRdmsDb, BSTR bsRdmsTbl, VARIANT vEventRecord, INT cfEventRecord, BOOL bHttpPost)
{
	// purpose = takes an event record and submits it using a xml/http get OR post to a sql2k database as an insert

	// note - when possible use xml http post, bHttpPost == TRUE, so we can avoid any issues that might arise if/when
	// event record data becomes large and thus our xml http get url string could end up exceeding the maximum
	// allowable url length for the xml http client libraries and or the http server receiving the request
	// see sql2k books on line "xmlAndInternetSupport / accessingSqlSvrUsingHttp / sampleAppsToPostTmplts / 
	// "Posting Templates Directly to the Virtual Directory" for a perfect example of how to easily port
	// xml http get routines to post routines, the Microsoft XML Parser SDK documentation found in the
	// download accessible on http://msdn.microsoft.com/xml -> xmlDownloads, the overview on sql2k updategrams
	// found on http://msdn.microsoft.com/downloads/betas/updategrams.asp, and finally see .\tests\test.vbp and
	// .\tests\*.xml for test examples of xml http get and post formats required 
	HRESULT hr = S_OK;
	BSTR bsTemp = NULL;
	try {
		// validate parameters
		if (!bsRdmsHost || !SysStringLen(bsRdmsHost) ||
			!bsRdmsDb || !SysStringLen(bsRdmsDb) ||
			!bsRdmsTbl || !SysStringLen(bsRdmsTbl) ||
			!vEventRecord.parray || !cfEventRecord) { hr = E_INVALIDARG; throw hr; }

		// note - for http post spaces are encoded with ' ', for http get spaces are supposed to be encoded with '+'
		// with '+'.  In testing it was found that http post will not succeed if spaces are encoded with '+' and
		// that http get will succeed with spaces encoded as ' ' therefore rather enabling support to encoding the
		// spaces in the xml differently depending on the value of bHttpPost we will just encode spaces with ' '
		// in both modes especially since it is expected that http post is the more optimal mode to operate given
		// you will not run into any issues of url lengths exceeding the client or servers maximum allowable length
		// _TCHAR gzSpcChar[] = _T(" "); if (!bHttpPost) _tcscpy(gzSpcChar, _T("+"));

		_TCHAR gzUrl[MAX_URL] = _T("\0");
		_tcscpy(gzUrl, _T("http://"));
		_tcscat(gzUrl, bsRdmsHost);
		_tcscat(gzUrl, _T("/"));
		_tcscat(gzUrl, bsRdmsDb);

		_TCHAR gzXml[MAX_XML] = _T("\0");
		// note - in the case of xml http get and post the gzXml should be prefixed by the xml version pi
		// and optionally the encoding setting but it appears that in the case of the msft xml parser it will
		// work even if it is missing
		// _tcscpy(gzXml, _T("<?xml version=\"1.0\" encoding=\"utf-8\"?>"));  // use ansii encoding
		// _tcscpy(gzXml, _T("<?xml version=\"1.0\" encoding=\"utf-16\"?>"));  // use unicode encoding
		_tcscpy(gzXml, _T("<?xml version=\"1.0\"?>"));  // use default encoding
		_tcscat(gzXml, _T("<root xmlns:updg=\"urn:schemas-microsoft-com:xml-updategram\">"));
		// or to support [xml-sql] queries as well
		// _tcscat(gzXml, _T("<root+xmlns:updg=\"urn:schemas-microsoft-com:xml-updategram\" "));
		// _tcscat(gzXml, _T("xmlns:sql=\"urn:schemas-microsoft-com:xml-sql\">"));
		_tcscat(gzXml, _T("<updg:sync>"));
		_tcscat(gzXml, _T("<updg:after>"));
		_tcscat(gzXml, _T("<"));
		_tcscat(gzXml, bsRdmsTbl);
		_tcscat(gzXml, _T(" "));
		SAFEARRAY *psaEventRecord = NULL;
		psaEventRecord = vEventRecord.parray;  // extract safearray pointer from variant array
		if (!psaEventRecord) { hr = HRESULT_FROM_WIN32(GetLastError()); throw hr; }

		LONG alCoords[2] = {0, 0};  // [0] == x coord;  [1] == y coord
		_variant_t vTemp;
		for (INT iX = 0; iX < cfEventRecord; iX++)
		{
			alCoords[0] = iX;  alCoords[1] = 0;  // [0] == x coord;  [1] == y coord
			HRESULT hr = SafeArrayGetElement(psaEventRecord, alCoords, &vTemp);  // column name
			if (FAILED(hr)) throw hr;
			vTemp.ChangeType(VT_BSTR, NULL);
			_tcscat(gzXml, vTemp.bstrVal);
			_tcscat(gzXml, _T("=\""));
			alCoords[0] = iX;  alCoords[1] = 1;  // [0] == x coord;  [1] == y coord
			hr = SafeArrayGetElement(psaEventRecord, alCoords, &vTemp);  // column value
			if (FAILED(hr)) throw hr;
			vTemp.ChangeType(VT_BSTR, NULL);
			_tcscat(gzXml, vTemp.bstrVal);
			_tcscat(gzXml, _T("\" "));
		}
		_tcscat(gzXml, _T("/>"));
		_tcscat(gzXml, _T("</updg:after>"));
		_tcscat(gzXml, _T("</updg:sync>"));
		/*  
		// useful for looking at different responseText but must be left out for product because it breaks our
		// responseText success string comparison later on, also requires that the root namespace value
		// xmlns:sql=\"urn:schemas-microsoft-com:xml-sql was included
		#if defined(_DEBUG)
			_tcscat(gzXml, _T("<sql:query> "));
			_tcscat(gzXml, _T("select * from "));
			_tcscat(gzXml, bsRdmsTbl);
			_tcscat(gzXml, _T(" order by iEventId for xml auto"));
			_tcscat(gzXml, _T("</sql:query>"));
		#endif
		*/
		_tcscat(gzXml, _T("</root>"));

		OutputDebugString(_T("\n")); OutputDebugString(gzUrl); OutputDebugString(_T("\n")); 
		OutputDebugString(gzXml); OutputDebugString(_T("\n")); 
		
		#if defined(_DEBUG)
			// validate our xml document using the msft xml parser
			IXMLDOMDocumentPtr spXMLDOMDocument;
			hr = spXMLDOMDocument.CreateInstance(CLSID_DOMDocument);  // class not registered if use __uuidof(IXMLDOMDocument));
			if (FAILED(hr)) throw hr;
			VARIANT_BOOL vbr = VARIANT_FALSE;
			hr = spXMLDOMDocument->loadXML(gzXml, &vbr);
			if (FAILED(hr)) throw hr;  
			// TODO: for some reason validate returns S_FALSE even though the xml payload is correct
			// as far as sql2k's xml parser is concerned and also if you load it up from file into ie55
		#endif

		// create and submit our xml http request
		IXMLHttpRequestPtr spXMLHttpRequest;
		hr = spXMLHttpRequest.CreateInstance(CLSID_XMLHTTPRequest, NULL);
		if (FAILED(hr)) throw hr;

		bool bAsync = FALSE;
		// _variant_t vAsync((bool)bAsync);  vAsync = VARIANT_FALSE;
		// send(...) method call used later is synchronous or asynchronous depending on the value of the bAsync 
		// parameter used here in the open call. If open is called with FALSE == bAsync, this call does not return
		// until the entire response is received or the protocol stack times out. If open is called with TRUE == bAsync
		// this call returns immediately [and the get_status(...), get_statusText(...), get_responseText(...) and
		// get_responseXML(...) method calls will be irrelevant].  It also appears from debug testing that
		// the xml technology preview for sql does not accept asynchronous submissions most likely because it leaves
		/// it unable to respond with success | failure result data.

		if (!bHttpPost) {
			_tcscat(gzUrl, _T("?template="));
			if ((sizeof(gzUrl) / sizeof(_TCHAR)) < _tcslen(gzUrl) + _tcslen(gzXml)) { hr = E_INVALIDARG; throw hr; }
			_tcscat(gzUrl, gzXml);
			OutputDebugString(_T("\n")); OutputDebugString(gzUrl); OutputDebugString(_T("\n")); 
		}

		_bstr_t sTemp;
		bHttpPost ? sTemp = _T("POST") : sTemp = _T("GET");
		hr = spXMLHttpRequest->open(sTemp, _bstr_t(gzUrl), _variant_t(bAsync), _variant_t(_T("")), _variant_t(_T("")));
		if (FAILED(hr)) throw hr;

		hr = spXMLHttpRequest->setRequestHeader(_bstr_t(_T("Content-Type")), _bstr_t(_T("application/xml")));  // | text/xml
		// if ::open(...) was called with bAsync == true, then this call returns immediately
		// which would perhaps be a big performance improvement under heavy loads
		// or if you spawned an i/o completion port thread to wait around for the
		// result or timeout to occur thus allowing this method to finish right away
		// and return
		if (FAILED(hr)) throw hr;

		vTemp.Clear();
		bHttpPost ? vTemp = _bstr_t(gzXml) : vTemp = _bstr_t(_T(""));
		hr = spXMLHttpRequest->send(vTemp);
		if (FAILED(hr)) {  // i.e. -2146697211/0x800C0005 == couldn't event talk to http server
			hr = WriteToLocalLogFile(_bstr_t(gzUrl), _bstr_t(gzXml), _bstr_t(_T(", ")));
			if (FAILED(hr)) throw hr;
		}
		else { // error check a little deeper into the http response details			
			LONG lTemp = 0L;
			hr = spXMLHttpRequest->get_status(&lTemp);
			if (E_UNEXPECTED != hr && (FAILED(hr) || (200 != lTemp))) {
				hr = WriteToLocalLogFile(_bstr_t(gzUrl), _bstr_t(gzXml), _bstr_t(_T(", ")));
				if (FAILED(hr)) throw hr;
			}
		}
		
		#if defined(_DEBUGFOO)  // futher error checking options
			hr = spXMLHttpRequest->get_statusText(&bsTemp);
			if (E_UNEXPECTED != hr)  {
				if (FAILED(hr) && (E_PENDING != hr)) throw hr;
				if (_tcscmp(_T("OK"), bsTemp)) { hr = E_FAIL; throw hr; }
			}

			if (bsTemp) { SysFreeString(bsTemp); bsTemp = NULL; }
			hr = spXMLHttpRequest->get_responseText(&bsTemp);
			if (FAILED(hr) && E_PENDING != hr) 
				throw hr;
			if (E_PENDING == hr)
				hr = S_OK;  // hr == E_PENDING if bAsync == TRUE, no way of telling if we succeeded at this time
			else {  // do xml payload failure detection 
				if (_tcscmp(_T("<?xml  version=\"1.0\" encoding=\"utf-8\"?><root xmlns:updg=\"urn:schemas-microsoft-com:xml-updategram\"></root>"), bsTemp)) {
					/* or if supporting  queries as well
					if (_tcscmp(_T("<?xml  version=\"1.0\" encoding=\"utf-8\"?><root xmlns:updg=\"urn:schemas-microsoft-com:xml-updategram\" xmlns:sql=\"urn:schemas-microsoft-com:xml-sql\"></root>"), bsTemp)) {
					*/
					// TODO: to add richer error recovery to handle this scenarios such as use of a registry key to denote
					// that submissions should be made to a msmq queue leaving it to a queue listner running on the 
					// centralized rdms host to guarantee record gets extracted and entered into centralized rdms database
					
					// note - if we throw an error the event will be resubmitted by wmi one more time giving us another shot
					// at getting it successfully logged which should help in cases of temporary network outages
					_bstr_t sTemp;
					sTemp = _T("received xml/http responseText value of ");
					sTemp += bsTemp;
					OutputDebugString(_T("\n")); OutputDebugString(sTemp); OutputDebugString(_T("\n")); 

					hr = S_FALSE; 
					throw hr;
				}
			}
		#endif
			
		#if defined(_FOOBAR)
			// try some access of IXMLDOMDocument testing how it can be walked to retrieve xml results info

			IXMLDOMDocumentPtr spXMLDOMDocument;
			hr = spXMLHttpRequest->get_responseXML((IDispatch **)&spXMLDOMDocument);
			if (E_PENDING != hr)  {  // i.e. if (!bAsync)
				if (FAILED(hr)) throw hr;

				IXMLDOMElementPtr spXMLDOMElement;
				hr = spXMLDOMDocument->get_documentElement(&spXMLDOMElement);
				if (FAILED(hr) || (NULL == spXMLDOMElement)) throw hr;

				// TODO:  how do you get your hands on elements not attributes in the domelement
				hr = spXMLDOMElement->getAttribute(_T("root"), &vTemp); 
				// if (FAILED(hr)) throw hr;

				if (S_FALSE != hr)  {  // temporary place holder until this approach for extended error checking is resolved
					// TODO:  add richer error recovery here we may want to add support for writing to a
					// locally stored csv circular log file so at a minimumm the event data is not lost...
					// and then notifying someone via another event mechanism or the fact that the csv
					// exists should be enough notification
					
					// if we throw an error the event will be resubmitted by wmi one more time giving
					// us another shot at getting it successfully logged in the cases where network
					// outages cause the failure to write
					throw hr;
				}
			}
		#endif
	}

	catch(...) {
		OutputDebugStringEx(hr, _bstr_t(_T("XmlHttpGetSqlInsert")));
		if (SUCCEEDED(hr)) throw;  // pass unexpected exceptions on to the outer block
	}

	if (bsTemp) {
		SysFreeString(bsTemp); 
		bsTemp = NULL;
	}

	return hr;
}


STDMETHODIMP CActions::XmlHttpSqlUpdate(BSTR bsRdmsHost, BSTR bsRdmsDb, BSTR bsRdmsTbl, VARIANT vEventRecord, INT cfEventRecord, BOOL bHttpPost)
{
	// purpose = takes an event record and submits it using a xml/http get OR post to a sql2k database as an update

	// note - when possible use xml http post, bHttpPost == TRUE, so we can avoid any issues that might arise if/when
	// event record data becomes large and thus our xml http get url string could end up exceeding the maximum
	// allowable url length for the xml http client libraries and or the http server receiving the request
	// see sql2k books on line "xmlAndInternetSupport / accessingSqlSvrUsingHttp / sampleAppsToPostTmplts / 
	// "Posting Templates Directly to the Virtual Directory" for a perfect example of how to easily port
	// xml http get routines to post routines, the Microsoft XML Parser SDK documentation found in the
	// download accessible on http://msdn.microsoft.com/xml -> xmlDownloads, the overview on sql2k updategrams
	// found on http://msdn.microsoft.com/downloads/betas/updategrams.asp, and finally see .\tests\test.vbp and
	// .\tests\*.xml for test examples of xml http get and post formats required 
	HRESULT hr = S_OK;
	BSTR bsTemp = NULL;
	try {
		// validate parameters
		if (!bsRdmsHost || !SysStringLen(bsRdmsHost) ||
			!bsRdmsDb || !SysStringLen(bsRdmsDb) ||
			!bsRdmsTbl || !SysStringLen(bsRdmsTbl) ||
			!vEventRecord.parray || !cfEventRecord) { hr = E_INVALIDARG; throw hr; }

		// note - for http post spaces are encoded with ' ', for http get spaces are supposed to be encoded with '+'
		// with '+'.  In testing it was found that http post will not succeed if spaces are encoded with '+' and
		// that http get will succeed with spaces encoded as ' ' therefore rather enabling support to encoding the
		// spaces in the xml differently depending on the value of bHttpPost we will just encode spaces with ' '
		// in both modes especially since it is expected that http post is the more optimal mode to operate given
		// you will not run into any issues of url lengths exceeding the client or servers maximum allowable length
		// _TCHAR gzSpcChar[] = _T(" "); if (!bHttpPost) _tcscpy(gzSpcChar, _T("+"));

		_TCHAR gzUrl[MAX_URL] = _T("\0");
		_tcscpy(gzUrl, _T("http://"));
		_tcscat(gzUrl, bsRdmsHost);
		_tcscat(gzUrl, _T("/"));
		_tcscat(gzUrl, bsRdmsDb);

		_TCHAR gzXml[MAX_XML] = _T("\0");
		// note - in the case of xml http get and post the gzXml should be prefixed by the xml version pi
		// and optionally the encoding setting but it appears that in the case of the msft xml parser it will
		// work even if it is missing
		// _tcscpy(gzXml, _T("<?xml version=\"1.0\" encoding=\"utf-8\"?>"));  // use ansii encoding
		// _tcscpy(gzXml, _T("<?xml version=\"1.0\" encoding=\"utf-16\"?>"));  // use unicode encoding
		_tcscpy(gzXml, _T("<?xml version=\"1.0\"?>"));  // use default encoding
		_tcscat(gzXml, _T("<root xmlns:updg=\"urn:schemas-microsoft-com:xml-updategram\">"));
		// or to support [xml-sql] queries as well
		// _tcscat(gzXml, _T("<root xmlns:updg=\"urn:schemas-microsoft-com:xml-updategram\" "));
		// _tcscat(gzXml, _T("xmlns:sql=\"urn:schemas-microsoft-com:xml-sql\">"));
		_tcscat(gzXml, _T("<updg:sync>"));
		_tcscat(gzXml, _T("<updg:before>"));
		// <updg:before> needs to write following columns at a minimum to locate existing record and to
		// include all the the fields that we will be updated in the updategram after clause
		// szHostname = 0 and/or szIpAddress = 1, szName = 3 (updated), szDescription = 4 (updated), dtCreate = 5
		// dtTerminate = 6 = "" (updated), iDuration = 7 = "" (updated), szOtherPropA = 9 == ProcessID,
		// [optional] szOtherPropB = 10, [optional] szOtherPropC = 11, etc.
		// note that the optionals are if only you want any additional cross references against szOtherPropA value 
		// for locating correct record when updategram is processed, see .\tests\xmlHttpPostSqlUpdate.xml for sample
		// updategram formats
		_tcscat(gzXml, _T("<"));
		_tcscat(gzXml, bsRdmsTbl);
		_tcscat(gzXml, _T(" "));
		SAFEARRAY *psaEventRecord = NULL;
		psaEventRecord = vEventRecord.parray;  // extract safearray pointer from variant array
		if (!psaEventRecord) { hr = HRESULT_FROM_WIN32(GetLastError()); throw hr; }

		LONG alCoords[2] = {0, 0};  // [0] == x coord;  [1] == y coord
		_variant_t vTemp;
		for (INT iX = 0; iX < cfEventRecord; iX++)
		{
			alCoords[0] = iX;  alCoords[1] = 0;  // [0] == x coord;  [1] == y coord
			HRESULT hr = SafeArrayGetElement(psaEventRecord, alCoords, &vTemp);  // column name
			if (FAILED(hr)) throw hr;
			vTemp.ChangeType(VT_BSTR, NULL);
			_tcscat(gzXml, vTemp.bstrVal);
			_tcscat(gzXml, _T("=\""));
			alCoords[0] = iX;  alCoords[1] = 1;  // [0] == x coord;  [1] == y coord
			hr = SafeArrayGetElement(psaEventRecord, alCoords, &vTemp);  // column value
			if (FAILED(hr)) throw hr;
			vTemp.ChangeType(VT_BSTR, NULL);
			switch (iX) {
				case 0:  // szHostname
					_tcscat(gzXml, vTemp.bstrVal);
					iX = 2; // manually bump it to number preceding next column we need in updategram
					break;
				case 3:  // szName
					if (!_tcscmp(PRC_TERMINATE_EVENT_NAME, vTemp.bstrVal))
						_tcscat(gzXml, PRC_CREATE_EVENT_NAME);
					else // i.e. GpepEvent where szName/Create event == szName/Terminate event
						_tcscat(gzXml, vTemp.bstrVal);
					break;
				case 4:  // szDescription
					if (!_tcscmp(PRC_TERMINATE_EVENT_DESCRIPTION, vTemp.bstrVal))
						_tcscat(gzXml, PRC_CREATE_EVENT_DESCRIPTION);
					else // i.e. GpepEvent where szName/Create event == szName/Terminate event
						_tcscat(gzXml, vTemp.bstrVal);
					break;
				case 6:  // dtTermination
					// add nothing so a NULL entry set to match the NULL entry set when Create event was inserted
					break;
				case 7:  // iDuration
					// add nothing so a NULL entry set to match the NULL entry set when Create event was inserted
					// iX = 8; // manually bump it to number preceding next column we need in updategram
					// don't need to because since we didn't send user record this time then everything
					// after column 7 shifted down one
					break;
				case 8:  // Win32_Process Create szOtherPropA == ProcessID, GpepEvent Create == Application
					_tcscat(gzXml, vTemp.bstrVal);
					iX = cfEventRecord; // manually bump it to number that will force for loop to exit
					break;
				default:
					_tcscat(gzXml, vTemp.bstrVal);
					break;
			}
			_tcscat(gzXml, _T("\" "));
		}
		_tcscat(gzXml, _T("/>"));
		_tcscat(gzXml, _T("</updg:before>"));
		
		_tcscat(gzXml, _T("<updg:after>"));
		// <updg:after> needs to write all the same columns as <updg:before> and this time include the updated
		// column values, see .\tests\xmlHttpPostSqlUpdate.xml for sample updategram formats
		_tcscat(gzXml, _T("<"));
		_tcscat(gzXml, bsRdmsTbl);
		_tcscat(gzXml, _T(" "));

		vTemp.Clear();
		for (iX = 0; iX < cfEventRecord; iX++)
		{
			alCoords[0] = iX;  alCoords[1] = 0;  // [0] == x coord;  [1] == y coord
			HRESULT hr = SafeArrayGetElement(psaEventRecord, alCoords, &vTemp);  // column name
			if (FAILED(hr)) throw hr;
			vTemp.ChangeType(VT_BSTR, NULL);
			_tcscat(gzXml, vTemp.bstrVal);
			_tcscat(gzXml, _T("=\""));
			alCoords[0] = iX;  alCoords[1] = 1;  // [0] == x coord;  [1] == y coord
			hr = SafeArrayGetElement(psaEventRecord, alCoords, &vTemp);  // column value
			if (FAILED(hr)) throw hr;
			vTemp.ChangeType(VT_BSTR, NULL);
			switch (iX) {
				case 0:  // szHostname
					_tcscat(gzXml, vTemp.bstrVal);
					iX = 2; // manually bump it to number preceding next column we need in updategram
					break;
				case 7:  // iDuration
					_tcscat(gzXml, vTemp.bstrVal);
					// iX = 8; // manually bump it to number preceding next column we need in updategram
					// don't need to because since we didn't send user record this time then everything
					// after column 7 shifted down one
					break;
				case 8: // szOtherPropA == ProcessID
					_tcscat(gzXml, vTemp.bstrVal);
					iX = cfEventRecord; // manually bump it to number that will force for loop to exit
					break;
				default:
					_tcscat(gzXml, vTemp.bstrVal);
					break;
			}
			_tcscat(gzXml, _T("\" "));
		}
		_tcscat(gzXml, _T("/>"));
		_tcscat(gzXml, _T("</updg:after>"));
		_tcscat(gzXml, _T("</updg:sync>"));
		/*  
		// useful for looking at different responseText but must be left out for product because it breaks our
		// responseText success string comparison later on, also requires that the root namespace value
		// xmlns:sql=\"urn:schemas-microsoft-com:xml-sql was included
		#if defined(_DEBUG)
			_tcscat(gzXml, _T("<sql:query> "));
			_tcscat(gzXml, _T("select * from "));
			_tcscat(gzXml, bsRdmsTbl);
			_tcscat(gzXml, _T(" order by iEventId for xml auto"));
			_tcscat(gzXml, _T("</sql:query>"));
		#endif
		*/
		_tcscat(gzXml, _T("</root>"));
		
		OutputDebugString(_T("\n")); OutputDebugString(gzUrl); OutputDebugString(_T("\n"));
		OutputDebugString(gzXml); OutputDebugString(_T("\n"));
		
		#if defined(_DEBUG)
			// validate our xml document using the msft xml parser
			IXMLDOMDocumentPtr spXMLDOMDocument;
			hr = spXMLDOMDocument.CreateInstance(CLSID_DOMDocument);  // class not registered if use __uuidof(IXMLDOMDocument));
			if (FAILED(hr)) throw hr;
			VARIANT_BOOL vbr = VARIANT_FALSE;
			hr = spXMLDOMDocument->loadXML(gzXml, &vbr);
			if (FAILED(hr)) throw hr;  
			// TODO: for some reason validate returns S_FALSE even though the xml payload is correct
			// as far as sql2k's xml parser is concerned and also if you load it up from file into ie55
		#endif

		// create and submit our xml http request
		IXMLHttpRequestPtr spXMLHttpRequest;
		hr = spXMLHttpRequest.CreateInstance(CLSID_XMLHTTPRequest, NULL);
		if (FAILED(hr)) throw hr;

		bool bAsync = FALSE;
		// _variant_t vAsync((bool)bAsync);  vAsync = VARIANT_FALSE;
		// send(...) method call used later is synchronous or asynchronous depending on the value of the bAsync 
		// parameter used here in the open call. If open is called with FALSE == bAsync, this call does not return
		// until the entire response is received or the protocol stack times out. If open is called with TRUE == bAsync
		// this call returns immediately [and the get_status(...), get_statusText(...), get_responseText(...) and
		// get_responseXML(...) method calls will be irrelevant].  It also appears from debug testing that
		// the xml technology preview for sql does not accept asynchronous submissions most likely because it leaves
		/// it unable to respond with success | failure result data.

		if (!bHttpPost) {
			_tcscat(gzUrl, _T("?template="));
			if ((sizeof(gzUrl) / sizeof(_TCHAR)) < _tcslen(gzUrl) + _tcslen(gzXml)) { hr = E_INVALIDARG; throw hr; }
			_tcscat(gzUrl, gzXml);
			OutputDebugString(_T("\n")); OutputDebugString(gzUrl); OutputDebugString(_T("\n")); 
		}

		_bstr_t sTemp;
		bHttpPost ? sTemp = _T("POST") : sTemp = _T("GET");
		hr = spXMLHttpRequest->open(sTemp, _bstr_t(gzUrl), _variant_t(bAsync), _variant_t(_T("")), _variant_t(_T("")));
		if (FAILED(hr)) throw hr;

		hr = spXMLHttpRequest->setRequestHeader(_bstr_t(_T("Content-Type")), _bstr_t(_T("application/xml")));  // | text/xml
		// if ::open(...) was called with bAsync == true, then this call returns immediately
		// which would perhaps be a big performance improvement under heavy loads
		// or if you spawned an i/o completion port thread to wait around for the
		// result or timeout to occur thus allowing this method to finish right away
		// and return
		if (FAILED(hr)) throw hr;

		vTemp.Clear();
		bHttpPost ? vTemp = _bstr_t(gzXml) : vTemp = _bstr_t(_T(""));
		hr = spXMLHttpRequest->send(vTemp);
		if (FAILED(hr)) {  // i.e. -2146697211/0x800C0005 == couldn't event talk to http server
			hr = WriteToLocalLogFile(_bstr_t(gzUrl), _bstr_t(gzXml), _bstr_t(_T(", ")));
			if (FAILED(hr)) throw hr;
		}
		else { // error check a little deeper into the http response details			
			LONG lTemp = 0L;
			hr = spXMLHttpRequest->get_status(&lTemp);
			if (E_UNEXPECTED != hr && (FAILED(hr) || (200 != lTemp))) {
				hr = WriteToLocalLogFile(_bstr_t(gzUrl), _bstr_t(gzXml), _bstr_t(_T(", ")));
				if (FAILED(hr)) throw hr;
			}
		}

		#if defined(_DEBUGFOO)  // futher error checking options
			hr = spXMLHttpRequest->get_statusText(&bsTemp);
			if (E_UNEXPECTED != hr)  {
				if (FAILED(hr) && (E_PENDING != hr)) throw hr;
				if (_tcscmp(_T("OK"), bsTemp)) { hr = E_FAIL; throw hr; }
			}

			if (bsTemp) { SysFreeString(bsTemp); bsTemp = NULL; }
			hr = spXMLHttpRequest->get_responseText(&bsTemp);
			if (FAILED(hr) && E_PENDING != hr) 
				throw hr;
			if (E_PENDING == hr)
				hr = S_OK;  // hr == E_PENDING if bAsync == TRUE, no way of telling if we succeeded at this time
			else {  // do xml payload failure detection 
				if (_tcscmp(_T("<?xml  version=\"1.0\" encoding=\"utf-8\"?><root xmlns:updg=\"urn:schemas-microsoft-com:xml-updategram\"></root>"), bsTemp)) {
					/* or if supporting  queries as well
					if (_tcscmp(_T("<?xml  version=\"1.0\" encoding=\"utf-8\"?><root xmlns:updg=\"urn:schemas-microsoft-com:xml-updategram\" xmlns:sql=\"urn:schemas-microsoft-com:xml-sql\"></root>"), bsTemp)) {
					*/
					// TODO: to add richer error recovery to handle this scenarios such as use of a registry key to denote
					// that submissions should be made to a msmq queue leaving it to a queue listner running on the 
					// centralized rdms host to guarantee record gets extracted and entered into centralized rdms database
					
					// note - if we throw an error the event will be resubmitted by wmi one more time giving us another shot
					// at getting it successfully logged which should help in cases of temporary network outages
					_bstr_t sTemp;
					sTemp = _T("received xml/http responseText value of ");
					sTemp += bsTemp;
					OutputDebugString(_T("\n")); OutputDebugString(sTemp); OutputDebugString(_T("\n"));

					hr = S_FALSE; 
					throw hr;
				}
			}
		#endif
			
		#if defined(_FOOBAR)
			// try some access of IXMLDOMDocument testing how it can be walked to retrieve xml results info

			IXMLDOMDocumentPtr spXMLDOMDocument;
			hr = spXMLHttpRequest->get_responseXML((IDispatch **)&spXMLDOMDocument);
			if (E_PENDING != hr)  {  // i.e. if (!bAsync)
				if (FAILED(hr)) throw hr;

				IXMLDOMElementPtr spXMLDOMElement;
				hr = spXMLDOMDocument->get_documentElement(&spXMLDOMElement);
				if (FAILED(hr) || (NULL == spXMLDOMElement)) throw hr;

				// TODO:  how do you get your hands on elements not attributes in the domelement
				hr = spXMLDOMElement->getAttribute(_T("root"), &vTemp); 
				// if (FAILED(hr)) throw hr;

				if (S_FALSE != hr)  {  // temporary place holder until this approach for extended error checking is resolved
					// TODO:  add richer error recovery here we may want to add support for writing to a
					// locally stored csv circular log file so at a minimumm the event data is not lost...
					// and then notifying someone via another event mechanism or the fact that the csv
					// exists should be enough notification
					
					// if we throw an error the event will be resubmitted by wmi one more time giving
					// us another shot at getting it successfully logged in the cases where network
					// outages cause the failure to write
					throw hr;
				}
			}
		#endif
	}

	catch(...) {
		OutputDebugStringEx(hr, _bstr_t(_T("XmlHttpGetSqlUpdate")));
		if (SUCCEEDED(hr)) throw;  // pass unexpected exceptions on to the outer block
	}

	if (bsTemp) {
		SysFreeString(bsTemp); 
		bsTemp = NULL;
	}

	return hr;
}
