// GnlPpsEvntCsmr.cpp : Implementation of DLL Exports.


// Note: Proxy/Stub Information
//      To build a separate proxy/stub DLL, 
//      run nmake -f GnlPpsEvntCsmrps.mk in the project directory.

#include "stdafx.h"
#include "resource.h"
#include <initguid.h>
#include "GnlPpsEvntCsmr.h"

#include "GnlPpsEvntCsmr_i.c"
#include "Wecp.h"
#include "Wuos.h"
#include "Actions.h"


CComModule _Module;

WSADATA _oWsaData = {0};
IFileSystemPtr _spFs;
ITextStreamPtr _spTxtStrm;
CRITICAL_SECTION _oLogFileCritSctn;
_bstr_t _sGpecDllPath;

BEGIN_OBJECT_MAP(ObjectMap)
OBJECT_ENTRY(CLSID_Wecp, CWecp)
OBJECT_ENTRY(CLSID_Wuos, CWuos)
OBJECT_ENTRY(CLSID_Actions, CActions)
END_OBJECT_MAP()

/////////////////////////////////////////////////////////////////////////////
// DLL Entry Point

extern "C"
BOOL WINAPI DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID /*lpReserved*/)
{
	HRESULT hr = S_OK;
    if (dwReason == DLL_PROCESS_ATTACH)
    {
		if (!_Module.GetLockCount()) {  
			// when first process attaches initialize globals
			
			// initialize the use of Ws2_32.dll by a process
			INT ir = WSAStartup(MAKEWORD(1, 1), &_oWsaData);
			if (ir) throw;  // TODO: comment out throw for vs.net build support



			/*
			// note - moved this initialization from here to "WriteToLocalLogFile" routine because it caused errors
			// when attempting to register component for debugging purposes under the com+ component services snapin
			// and in addition there is no sense creating a log file on the disk until such time that a network or
			// rdms down scenario forces us to write out event data to a local log file instead

			// initialize file system object that will be used by WriteToLocalLogFile calls
			// when errors occur attempting to write event records to the centralized rdms
			// hr = _spFs.CreateInstance(__uuidof(IFileSystem), NULL);  // don't use returns hr == 0x80040154 Class not registered
			hr = _spFs.CreateInstance(_T("Scripting.FileSystemObject")), NULL);
			if (FAILED(hr)) throw hr;  // TODO: comment out throw for vs.net build support
			*/

			// initialize the log file resource critical section object
			InitializeCriticalSection(&_oLogFileCritSctn);
			hr = HRESULT_FROM_WIN32(GetLastError());
			if (FAILED(hr)) throw;  // TODO: comment out throw for vs.net build support
		}

		// _Module.Init(ObjectMap, hInstance, &LIBID_GNLPPSEVNTCSMRLib);
        _Module.Init(ObjectMap, hInstance, &LIBID_GnlPpsEvntCsmr);
        DisableThreadLibraryCalls(hInstance);
		#if (_DEBUG)
			INT iLockCount = _Module.GetLockCount();
		#endif
    }
    else {
		if (dwReason == DLL_PROCESS_DETACH) {
			if (!_Module.GetLockCount()) {  
				// when last process detaches uninitialize globals
				
				// uninitialize the use of Ws2_32.dll by a process
				INT ir = WSACleanup();
				if (ir) throw;  // TODO: comment out throw for vs.net build support

				// release the resources used by the log file resource critical section object.
				DeleteCriticalSection(&_oLogFileCritSctn);
				hr = HRESULT_FROM_WIN32(GetLastError());
				if (FAILED(hr)) throw;  // TODO: comment out throw for vs.net build support
			}

			_Module.Term();
		}
	}

    return TRUE;    // ok
}

/////////////////////////////////////////////////////////////////////////////
// Used to determine whether the DLL can be unloaded by OLE

STDAPI DllCanUnloadNow(void)
{
    return (_Module.GetLockCount()==0) ? S_OK : S_FALSE;
}

/////////////////////////////////////////////////////////////////////////////
// Returns a class factory to create an object of the requested type

STDAPI DllGetClassObject(REFCLSID rclsid, REFIID riid, LPVOID* ppv)
{
    return _Module.GetClassObject(rclsid, riid, ppv);
}

/////////////////////////////////////////////////////////////////////////////
// DllRegisterServer - Adds entries to the system registry

STDAPI DllRegisterServer(void)
{
    // registers object, typelib and all interfaces in typelib
    return _Module.RegisterServer(TRUE);
}

/////////////////////////////////////////////////////////////////////////////
// DllUnregisterServer - Removes entries from the system registry

STDAPI DllUnregisterServer(void)
{
    return _Module.UnregisterServer(TRUE);
}


double _ttof(const _TCHAR *string)
{
	// purpose = convert a generic text string to a float, suppliments the missing cruntime generic text version
	// of atof and as is the case with atof the return value is 0.0 if the input cannot be converted and is undefined
	// in the case of overflow.  The string parameter is expected to be a representation of float or a double with a 
	// format of [whitespace] [sign] [digits] [.digits]
	#if defined(_UNICODE)
		char szTemp[MAX_FLOAT] = "\0";
		size_t str = wcstombs(szTemp, string, _tcslen(string));
		if (!str) return 0;  
		return atof(szTemp);  
	#else
		return atof(string);
	#endif
}


HRESULT OutputDebugStringEx(HRESULT hrError, BSTR bsRoutine)
{
	// purpose = extract a string representation of HRESULT and send to output window to support debugging
	// on NDEBUG builds function does nothing
	HRESULT hr = S_OK;

#if defined(_DEBUG)
	LPTSTR pgzTemp = NULL;

	try {
		// validate parameters
		if (/* !hrErrorInfo || */ !bsRoutine || !SysStringLen(bsRoutine)) { hr = E_INVALIDARG; throw hr; }

		DWORD dwr = FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_SYSTEM,
								  NULL, hrError, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR)&pgzTemp, 0, NULL);
		if (!dwr && (ERROR_MR_MID_NOT_FOUND != GetLastError())) { hr = HRESULT_FROM_WIN32(GetLastError()); throw hr; }
		_bstr_t sTemp;
		sTemp = _T("Error processing \"");
		sTemp += bsRoutine;
		sTemp += _T("\" resulting in a FormatMessage description = ");
		if (pgzTemp && _tcslen(pgzTemp))
			sTemp += pgzTemp;
		else
			sTemp += _T("message not found");
		OutputDebugString(sTemp);
	}
	
	catch(...) {
		// do nothing in the case of a expected exceptions for this function
		if (SUCCEEDED(hr)) throw;  // pass unexpected exceptions on to the outer block
	}

	IErrorInfo *pErrorInfo = NULL;
	HANDLE hHeap = GetProcessHeap();
	if (!hHeap || (INVALID_HANDLE_VALUE == hHeap)) throw;
	BOOL br = HeapFree(hHeap, 0, pgzTemp);
	if (!br) throw;
	pgzTemp = NULL;
#endif

	return hr;
}


HRESULT GetSystemTimeEx(BSTR bsWmiDatetime, LPSYSTEMTIME pSystemTime)
{
	// purpose = convert a wmi datetime value to a SYSTEMTIME structure
	// wmi datetime format = yyyymmddHHMMSS.mmmmmmsUUU, ranging from January 1, 1601 onward
	// all fields must be of the indicated length, using leading zeros as appropriate for the type. 
	// mmmmmm = six digit representation of microseconds
	// s = + | - offset from UTC (GMT has UTC offset of 0 seconds)
	// UUU = three digit representation of the number of minutes that originating time zone deviates from UTC
	// for WMI, it is encouraged, but not required, that you convert times to GMT with a UTC offset of zero.
	// SystemTime structure is always provided in GMT or Coordinated Universal Time (UTC). 
	HRESULT hr = S_OK;
	try {
		// validate parameters
		if (!bsWmiDatetime || !SysStringLen(bsWmiDatetime) || !pSystemTime) { hr = E_INVALIDARG; throw hr; }

		// ensure the structure is empty before use
		memset(pSystemTime, 0, sizeof(SYSTEMTIME));

		_TCHAR gzTemp[WMIDATETIME_MAX_FIELD] = _T("\0");
		_tcsncpy(gzTemp, bsWmiDatetime + WMIDATETIME_INDEXOF_YYYY, (WMIDATETIME_INDEXOF_MM - WMIDATETIME_INDEXOF_YYYY));
		pSystemTime->wYear  = _ttoi(gzTemp);

		_tcsset(gzTemp, _T('\0'));
		_tcsncpy(gzTemp, bsWmiDatetime + WMIDATETIME_INDEXOF_MM, (WMIDATETIME_INDEXOF_DD - WMIDATETIME_INDEXOF_MM));
		pSystemTime->wMonth = _ttoi(gzTemp);

		_tcsset(gzTemp, _T('\0'));
		_tcsncpy(gzTemp, bsWmiDatetime + WMIDATETIME_INDEXOF_DD, (WMIDATETIME_INDEXOF_HH - WMIDATETIME_INDEXOF_DD));
		pSystemTime->wDay = _ttoi(gzTemp);

		_tcsset(gzTemp, _T('\0'));
		_tcsncpy(gzTemp, bsWmiDatetime + WMIDATETIME_INDEXOF_HH, (WMIDATETIME_INDEXOF_MIN - WMIDATETIME_INDEXOF_HH));
		INT iUtcOffset = _ttoi(bsWmiDatetime + WMIDATETIME_INDEXOF_UTCOFFSET) / 60;
		INT iHour = _ttoi(gzTemp) - iUtcOffset;
		if (24 <= iHour) {  // for events that happen to the west of GMT equating to the next day in GMT
			iHour -= 24;
			++pSystemTime->wDay;
			
		}
		if (0 > iHour) {  // for events that happen to the east of GMT equating to the previous day in GMT
			iHour += 24;
			--pSystemTime->wDay;
		}
		
		pSystemTime->wHour = iHour;

		_tcsset(gzTemp, _T('\0'));
		_tcsncpy(gzTemp, bsWmiDatetime + WMIDATETIME_INDEXOF_MIN, (WMIDATETIME_INDEXOF_SS - WMIDATETIME_INDEXOF_MIN));
		pSystemTime->wMinute = _ttoi(gzTemp);

		_tcsset(gzTemp, _T('\0'));
		_tcsncpy(gzTemp, bsWmiDatetime + WMIDATETIME_INDEXOF_SS, (WMIDATETIME_INDEXOF_MS - WMIDATETIME_INDEXOF_SS - 1));  // - 1 accounts for '.' character
		pSystemTime->wSecond = _ttoi(gzTemp);

		_tcsset(gzTemp, _T('\0'));
		_tcsncpy(gzTemp, bsWmiDatetime + WMIDATETIME_INDEXOF_MS, (WMIDATETIME_INDEXOF_UTCOFFSET - WMIDATETIME_INDEXOF_MS));
		LPTSTR pgzTemp = NULL;
		#if defined(_DEBUG)
			int iTemp = INT(_tcstod(gzTemp, &pgzTemp) / 1000 + 0.5);
			iTemp = INT(_tcstod(_T("203000"), &pgzTemp) / 1000 + 0.5);  // round down, 203.500
			iTemp = INT(_tcstod(_T("203001"), &pgzTemp) / 1000 + 0.5);  // round down, 203.501
			iTemp = INT(_tcstod(_T("203499"), &pgzTemp) / 1000 + 0.5);  // round down, 203.999
			iTemp = INT(_tcstod(_T("203500"), &pgzTemp) / 1000 + 0.5);  // round up, 204.000
			iTemp = INT(_tcstod(_T("203501"), &pgzTemp) / 1000 + 0.5);  // round up, 204.001
			iTemp = INT(_tcstod(_T("203999"), &pgzTemp) / 1000 + 0.5);  // round up, 204.499
			iTemp = INT(_tcstod(_T("204000"), &pgzTemp) / 1000 + 0.5);  // round up, 204.500
			pgzTemp = NULL;
		#endif
		DOUBLE dTemp = _tcstod(gzTemp, &pgzTemp);
		if (_tcscmp(_T("\0"), pgzTemp)) { pgzTemp = NULL; hr = E_INVALIDARG; throw hr; }  // a 100% conversion didn't occur
		pSystemTime->wMilliseconds = INT(dTemp / 1000 + 0.5);  
		// division by 1000 does micro to milliseconds conversion and addition of .5 enables round off
	}
	
	catch(...) {
		OutputDebugStringEx(hr, _bstr_t(_T("GetSystemTimeEx")));
		if (SUCCEEDED(hr)) throw;  // pass unexpected exceptions on to the outer block
	}

	return hr;
}


HRESULT GetSystemTimeElapsedInSeconds(SYSTEMTIME &oCreateSystemTime, SYSTEMTIME &oTerminateSystemTime, LONG *plElapsedSeconds)
{
	// purpose = calculated the absolute elapsed time in seconds between two different SYSTEMTIME structs
	HRESULT hr = S_OK;
	try {
		// validate parameters
		if (/* !oCreateSystemTime */ !oCreateSystemTime.wYear || !oTerminateSystemTime.wYear || !plElapsedSeconds) { hr = E_INVALIDARG; throw hr; }

		/*
		// option 1 = mfc class wrapper for time data and functions
		// note - requires inclusion of afx.h will generates compile error if project already includes windows.h
		COleDateTime oCreateDateTime(oCreateSystemTime), oTerminateDateTime(oTerminateSystemTime);
		COleDateTimeSpan oDateTimeSpan = oTerminateDateTime - oCreateDateTime;
		// _variant_t vElapsedMinutes = oDateTimeSpan.GetTotalMinutes();
		_variant_t vElapsedSeconds = oDateTimeSpan.GetTotalSeconds();
		vElapsedSeconds.ChangeType(VT_I4, NULL);
		*plElapsedSeconds = vElapsedSeconds.lVal;
		*/

		/*
		// option 2 = #include <afx.h> mfc class wrapper derivation of ansi time_t data type and its associated run-time functions
		// note - requires inclusion of afx.h will generates compile error if project already includes windows.h
		CTime oCreateTime(oCreateSystemTime, 0), oTerminateTime(oTerminateSystemTime, 0);
		CTimeSpan oTimeSpan = oTerminateTime - oCreateTime;
		// *plElapsedMinutes = oTimeSpan.GetTotalMinutes();
		*plElapsedSeconds = oTimeSpan.GetTotalSeconds();
		*/

		/*
		// option 3 - include <time.h> ansi time_t data type and its associated run-time functions
		// TODO: use time_t variables which represent number of seconds elapsed since midnight (00:00:00), 
		// January 1, 1970, coordinated universal timesince and some routine to convert SYSTEMTIME values to 
		// this format, then do simple subtraction to get time span and if necessary to return value in minutes
		// then divide by 60, round off, and return for number of minutes elapsed
		struct tm oCreateTime = {0}, oTerminateTime = {0}
		oCreateTimeM.tm_year = oCreateSystemTime.wYear; oCreateTimeM.tm_mon = oCreateSystemTime.wMonth; 
		oCreateTimeM.tm_mday = oCreateSystemTime.wDay; oCreateTimeM.tm_hour = oCreateSystemTime.wHour;
		oCreateTimeM.tm_min = oCreateSystemTime.wMinute; oCreateTimeM.tm_sec = oCreateSystemTime.wSecond;

		oTerminateTimeM.tm_year = oTerminateSystemTime.wYear; oTerminateTimeM.tm_mon = oTerminateSystemTime.wMonth; 
		oTerminateTimeM.tm_mday = oTerminateSystemTime.wDay; oTerminateTimeM.tm_hour = oTerminateSystemTime.wHour;
		oTerminateTimeM.tm_min = oTerminateSystemTime.wMinute; oTerminateTimeM.tm_sec = oTerminateSystemTime.wSecond;

		// this is where trouble arises since there is no way to convert struct tm values to absolute time_t values
		// in order to execute a difftime(oEnd, oStart) so as to represent the difference in absolute minutes
		// having to author something to convert struct tm values to absolute time_t values puts us right back at
		// square one, in addition to the fact that we lost our millisecond accurracy in the conversion of
		// a SYSTEMTIME to a struct tm above
		time_t oCreateTimeT = SomeStructTmToTimeTConvertor(oCreateTimeM),
			   oTerminateTimeT = SomeStructTmToTimeTConvertor(oTerminateTimeM);
		*plElapsedSeconds = (LONG)difftime(oTerminateTimeT, oCreateTimeT);
		*/

		// option 4 - as per the psdk documents just convert SYSTEMTIME structured date representation to a 
		// to FILETIME absolute time representation in nanoseconds and then do difference calculations and
		// do absolute difference in seconds calculation work from there
		FILETIME oCreateFileTime = {0}, oTerminateFileTime = {0};
		BOOL br = SystemTimeToFileTime(&oCreateSystemTime, &oCreateFileTime);
		if (!br) { hr = HRESULT_FROM_WIN32(GetLastError()); throw; }
		br = SystemTimeToFileTime(&oTerminateSystemTime, &oTerminateFileTime);
		if (!br) { hr = HRESULT_FROM_WIN32(GetLastError()); throw; }
		ULARGE_INTEGER uliCreateFileTime ={0}, uliTerminateFileTime={0}, uliElapsedSeconds={0};

		uliCreateFileTime.LowPart = oCreateFileTime.dwLowDateTime;
		uliCreateFileTime.HighPart = oCreateFileTime.dwHighDateTime;

		uliTerminateFileTime.LowPart = oTerminateFileTime.dwLowDateTime;
		uliTerminateFileTime.HighPart = oTerminateFileTime.dwHighDateTime;

		#if defined(_DEBUG)
			__int64 uli = (__int64)pow(10, 7);
			uliElapsedSeconds.QuadPart = (uliTerminateFileTime.QuadPart - uliCreateFileTime.QuadPart) / (__int64)10000000;
		#endif
			
		uliElapsedSeconds.QuadPart = (uliTerminateFileTime.QuadPart - uliCreateFileTime.QuadPart) / (__int64)pow(10, 7);
		/*
			#include <math.h> causes following linker error
		
			Linking...
			Creating library ReleaseUMinSize/GnlPpsEvntCsmr.lib and object ReleaseUMinSize/GnlPpsEvntCsmr.exp
			LIBCMT.lib(crt0.obj) : error LNK2001: unresolved external symbol _main
			ReleaseUMinSize/GnlPpsEvntCsmr.dll : fatal error LNK1120: 1 unresolved externals
			Error executing link.exe.
			GnlPpsEvntCsmr.dll - 2 error(s), 0 warning(s)
				
			q167946 fix - Remove _ATL_MIN_CRT from the list the preprocessor defines to allow CRT startup code to
			be included. 
		*/

		if (0 == uliElapsedSeconds.HighPart)
			*plElapsedSeconds = uliElapsedSeconds.LowPart;
		else {  // time difference is to big to return in a LONG variable so set to zero and throw error
			*plElapsedSeconds = 0;
			hr = E_FAIL; throw hr;
		}
	}
	
	catch(...) {
		OutputDebugStringEx(hr, _bstr_t(_T("GetSystemTimeElapsedInSeconds")));
		if (SUCCEEDED(hr)) throw;  // pass unexpected exceptions on to the outer block
	}

	return hr;
}


HRESULT GetSqlDatetime(SYSTEMTIME &oSystemTime, BSTR *pbsSqlDatetime)
{
	// purpose = convert a SYSTEMTIME structure to a sql datetime value
	// sql datetime format = yyyy-mm-dd hh:mm:ss.mmm, ranging from January 1, 1753 through December 31, 9999
	// mmm  = milliseconds rounded by sql on when stored to nearest 3.33 millisecond
	HRESULT hr = S_OK;
	try {
		// validate parameters
		if (/* !oSystemTime */ !oSystemTime.wYear || !pbsSqlDatetime) { hr = E_INVALIDARG; throw hr; }

		_TCHAR gzTemp[MAX_SQLDATETIME] = _T("\0"),
			   gzItotBuff[MAX_SYSTEMTIME_FIELD] = _T("\0");

		_tcscpy(gzTemp, _itot(oSystemTime.wYear, gzItotBuff, 10));
		_tcscat(gzTemp, _T("-"));
		
		if (10 > oSystemTime.wMonth)
			_tcscat(gzTemp, _T("0"));
		_tcscat(gzTemp, _itot(oSystemTime.wMonth, gzItotBuff, 10));
		_tcscat(gzTemp, _T("-"));
		
		if (10 > oSystemTime.wDay)
			_tcscat(gzTemp, _T("0"));
		_tcscat(gzTemp, _itot(oSystemTime.wDay, gzItotBuff, 10));
		_tcscat(gzTemp, _T(" "));
		
		if (10 > oSystemTime.wHour)
			_tcscat(gzTemp, _T("0"));
		_tcscat(gzTemp, _itot(oSystemTime.wHour, gzItotBuff, 10));
		_tcscat(gzTemp, _T(":"));
		
		if (10 > oSystemTime.wMinute)
			_tcscat(gzTemp, _T("0"));
		_tcscat(gzTemp, _itot(oSystemTime.wMinute, gzItotBuff, 10));
		_tcscat(gzTemp, _T(":"));
		
		if (10 > oSystemTime.wSecond)
			_tcscat(gzTemp, _T("0"));
		_tcscat(gzTemp, _itot(oSystemTime.wSecond, gzItotBuff, 10));
		_tcscat(gzTemp, _T("."));
		
		if (100 > oSystemTime.wMilliseconds)
			_tcscat(gzTemp, _T("0"));
		if (10 > oSystemTime.wMilliseconds)
			_tcscat(gzTemp, _T("0"));
		_tcscat(gzTemp, _itot(oSystemTime.wMilliseconds, gzItotBuff, 10));

		INT ir = 0;
		#if defined(_UNICODE)
			ir = SysReAllocString(pbsSqlDatetime, gzTemp);
		#else  // _MBCS
			WCHAR uzTemp[MAX_SQLDATETIME] = _T("\0");
			size_t str = mbstowcs(uzTemp, (LPCSTR)gzTemp, _tcslen(gzTemp));
			ir = SysReAllocString(pbsSqlDatetime, uzTemp);
		#endif
		if (!ir) { hr = HRESULT_FROM_WIN32(GetLastError()); throw hr; }
	}
	
	catch(...) {
		OutputDebugStringEx(hr, _bstr_t(_T("GetSqlDatetime")));
		if (SUCCEEDED(hr)) throw;  // pass unexpected exceptions on to the outer block
	}

	return hr;
}


HRESULT GetYearMonthDay(SYSTEMTIME &oSystemTime, BSTR *pbsYearMonthDay)
{
	// purpose = convert a SYSTEMTIME structure to a yyyy-mm-dd datetime value

	HRESULT hr = S_OK;

	try {
		// validate parameters
		if (/* !oSystemTime */ !oSystemTime.wYear || !pbsYearMonthDay) { hr = E_INVALIDARG; throw hr; }

		_TCHAR gzTemp[MAX_SQLDATETIME] = _T("\0"),
			   gzItotBuff[MAX_SYSTEMTIME_FIELD] = _T("\0");

		_tcscpy(gzTemp, _itot(oSystemTime.wYear, gzItotBuff, 10));
		_tcscat(gzTemp, _T("-"));
		
		if (10 > oSystemTime.wMonth)
			_tcscat(gzTemp, _T("0"));
		_tcscat(gzTemp, _itot(oSystemTime.wMonth, gzItotBuff, 10));
		_tcscat(gzTemp, _T("-"));
		
		if (10 > oSystemTime.wDay)
			_tcscat(gzTemp, _T("0"));
		_tcscat(gzTemp, _itot(oSystemTime.wDay, gzItotBuff, 10));

		INT ir = 0;
		#if defined(_UNICODE)
			ir = SysReAllocString(pbsYearMonthDay, gzTemp);
		#else  // _MBCS
			WCHAR uzTemp[MAX_SQLDATETIME] = _T("\0");
			size_t str = mbstowcs(uzTemp, (LPCSTR)gzTemp, _tcslen(gzTemp));
			ir = SysReAllocString(pbsYearMonthDay, uzTemp);
		#endif
		if (!ir) { hr = HRESULT_FROM_WIN32(GetLastError()); throw hr; }
	}
	
	catch(...) {
		OutputDebugStringEx(hr, _bstr_t(_T("GetYearMonthDay")));
		if (SUCCEEDED(hr)) throw;  // pass unexpected exceptions on to the outer block
	}

	return hr;
}


HRESULT GetOwnerWmi(BSTR bsNamespace, BSTR bsClassPath, BSTR bsInstancePath, BSTR *pbsSamAccountName, BSTR *pbsNetbiosDomain)
{
	// purpose = call the Win32_Process class GetOwner method on the passed Win32_Process instance that was
	// passed to this routine to retrieve the user and domain values associated with that instance
	HRESULT hr = S_OK;
	BSTR bsTemp = NULL;

	try {
		// validate parameters
		if (!bsNamespace || !SysStringLen(bsNamespace) || !bsClassPath || !SysStringLen(bsClassPath) ||
			!bsInstancePath || !SysStringLen(bsInstancePath) ||!pbsSamAccountName || !pbsNetbiosDomain) { hr = E_INVALIDARG; throw hr; }

		// example [in] parameters are something such as the following
		// bsNamespace = "root\cimv2\msftNsg" ||
		//			     "\\.\root\cimv2\msftNsg"
		// bsClassPath = "\\MYDEVB\root\cimv2\msftNsg:ViewClassUnionOfRootCimv2Win32Process" ||
		//				 "ViewClassUnionOfRootCimv2Win32Process"
		// bsInstancePath = "\\MYDEVB\root\cimv2\msftNsg:ViewClassUnionOfRootCimv2Win32Process.Handle="<handleValue>"" ||
		//                   "ViewClassUnionOfRootCimv2Win32Process.Handle="<handleValue>""

		// connect up to cimom
		IWbemLocatorPtr spLocator;
		// psdk way
		// hr = CoCreateInstance(CLSID_WbemLocator, 0, CLSCTX_INPROC_SERVER, IID_IWbemLocator, (LPVOID *)&spLocator);
		// smartpointer way
		hr = spLocator.CreateInstance(CLSID_WbemLocator);  // class not registered error if use (__uuidof(IWbemLocator));
		if (FAILED(hr)) throw hr;

		// connect to namespace where Win32_Process class instance withthe Win32_Process::GetOwner method exists
		IWbemServicesPtr spServices;
		hr = spLocator->ConnectServer(bsNamespace, NULL, NULL, NULL, 0, NULL, NULL, &spServices);
		if (FAILED(hr)) throw hr;

		// get the specific Win32_Process class we want to call the Win32_Process::GetOwner method on
		IWbemClassObjectPtr spClass;
		hr = spServices->GetObject(bsClassPath, 0, NULL, &spClass, NULL);
		if (FAILED(hr)) throw hr;

		// get the specific Win32_Process class instance we want to call the Win32_Process::GetOwner method on
		IWbemClassObjectPtr spInstance;
		hr = spServices->GetObject(bsInstancePath, 0, NULL, &spInstance, NULL);
		if (FAILED(hr)) throw hr;
		
		// the calls to acquire the method input parameters class and spawn and instance of it result in 
		// the errors noted depending one whether we use the class or an instance of it to attempt to
		// get the input parameters class, not sure why, possibly because this method only has [out] params
		// compare method flow with psdk samples for ExecMethod calls and anything of this nature on
		// http://wmig samples as well.  If we can't make it work right now then we will just use the
		// GetOwnerPsdk method to acquired the process owner and swap back in the native wmi approach
		// once we can get someone to have a look at this and comment on where you are going wrong

		// get the Win32_Process::GetOwner method input parameters class
		IWbemClassObjectPtr spInParamsClass;
		// hr = spClass->GetMethod(_bstr_t(_T("GetOwner")), 0, &spInParamsClass, NULL); 
		// spClass->GetMethod returns hr = 0x80041002 on some occassions, hr value has an unknown formatmessage description
		hr = spInstance->GetMethod(_bstr_t(_T("GetOwner")), 0, &spInParamsClass, NULL); 
		// spInstance->GetMethod returns hr = 0x8004101e on some occassions, hr value has an unknown formatmessage description
		if (FAILED(hr)) throw hr;  
	
		// spawn a Win32_Process::GetOwner method input parameters class instance
		IWbemClassObjectPtr spInParamsInstance;
		hr = spInParamsClass->SpawnInstance(0, &spInParamsInstance);  // this throws and exception
		if (FAILED(hr)) throw hr;

		// set the Win32_Process::GetOwner method input parameters
		_variant_t vTemp; vTemp.pbstrVal = pbsSamAccountName;
		hr = spInParamsInstance->Put(_bstr_t(_T("User")), 0, &vTemp, 0);
		if (FAILED(hr)) throw hr;

		vTemp.Clear(); vTemp.pbstrVal = pbsNetbiosDomain;
		hr = spInParamsInstance->Put(_bstr_t(_T("Domain")), 0, &vTemp, 0);
		if (FAILED(hr)) throw hr;

		// execute the Win32_Process::GetOwner method on the desired Win32_Process class instance
		// note that the format of relpath = ViewClassUnionOfWin32Process.Handle="<handleValue>"
		IWbemClassObjectPtr spOutParamsInstance;
		hr = spServices->ExecMethod(bsInstancePath, _bstr_t(_T("GetOwner")), 0, NULL, NULL, &spOutParamsInstance, NULL);
		if (FAILED(hr)) throw hr;

		// obtain [if necessary] the Win32_Process::GetOwner method call output parameters
		// this method call is supposed to return output parameters in mof text representation
		// but our origional BSTR * variables should already hold what we are looking for...I think
		hr = spOutParamsInstance->GetObjectText(0, &bsTemp);
		if (FAILED(hr)) throw hr;
	}
	
	catch(...) {
		OutputDebugStringEx(hr, _bstr_t(_T("GetOwnerWmi")));
		if (SUCCEEDED(hr)) throw;  // pass unexpected exceptions on to the outer block
	}

	if (bsTemp) { SysFreeString(bsTemp); bsTemp = NULL; }

	return hr;
}


HRESULT GetOwnerWin32(DWORD dwProcessId, BSTR *pbsSamAccountName, BSTR *pbsNetbiosDomain)
{
	// purpose = call the Win32_Process class GetOwner method on the passed Win32_Process instance that was
	// passed to this routine to retrieve the user and domain values associated with that instance
	HRESULT hr = S_OK;
	HANDLE hProcess = NULL, hProcessToken = NULL;

	try {
		// validate parameters
		if (!dwProcessId || !pbsSamAccountName || !pbsNetbiosDomain) { hr = E_INVALIDARG; throw hr; }

		// get caller and current process handles for use in duplicate handle steps
		hProcess = OpenProcess(PROCESS_QUERY_INFORMATION, FALSE, dwProcessId);
		if (!hProcess || (INVALID_HANDLE_VALUE == hProcess)) { hr = HRESULT_FROM_WIN32(GetLastError()); throw hr; }

		BOOL br = OpenProcessToken(hProcess, TOKEN_QUERY, &hProcessToken);
		if (!br) { hr = HRESULT_FROM_WIN32(GetLastError()); throw hr; }

		// get principal name from token
        BYTE abyTokInfo[MAX_TOKENINFO];  // if return type is a random size
		// BYTE abyTokInfo[sizeof(TOKEN_USER)];  // if return type is a known size
		#if defined(_DEBUG)
			INT soTu = sizeof(TOKEN_USER);
		#endif
		DWORD soTokInfo = sizeof(abyTokInfo), soReqd = 0;
		br = GetTokenInformation(hProcessToken, TokenUser, abyTokInfo, soTokInfo, &soReqd);
		if (!br) { hr = HRESULT_FROM_WIN32(GetLastError()); throw hr; }

		// PSID sidPrincipal = reinterpret_cast<SID*>(reinterpret_cast<TOKEN_USER*>(abyTokInfo)->User.Sid);
		// PSID psidPrincipal = PSID(PTOKEN_USER(abyTokInfo)->User.Sid);
		PSID psidPrincipal = PTOKEN_USER(abyTokInfo)->User.Sid;

		// extract the account name
		_TCHAR gzAuthority[MAX_AUTHORITY], gzPrincipal[MAX_PRINCIPAL];
		DWORD soAuthority = sizeof(gzAuthority), soPrincipal = sizeof(gzPrincipal);
		#if defined(_DEBUG)
			DWORD ccTemp = sizeof(gzAuthority) / sizeof(_TCHAR);
			ccTemp = sizeof(gzPrincipal) / sizeof(_TCHAR);
		#endif
		SID_NAME_USE eSidNameUse;
		br = LookupAccountSid(NULL, psidPrincipal, gzPrincipal, &soPrincipal, gzAuthority, &soAuthority, &eSidNameUse);
		if (!br) { hr = HRESULT_FROM_WIN32(GetLastError()); throw hr; }
			
		#if defined(_UNICODE)
			INT ir = SysReAllocString(pbsSamAccountName, gzPrincipal);
			if (!ir) { hr = HRESULT_FROM_WIN32(GetLastError()); throw hr; }
			ir = SysReAllocString(pbsNetbiosDomain, gzAuthority);
			if (!ir) { hr = HRESULT_FROM_WIN32(GetLastError()); throw hr; }
		#else  // _MBCS
			WCHAR uzTemp[MAX_AUTHORITY > MAX_PRINCIPAL ? MAX_AUTHORITY : MAX_PRINCIPAL] = _T("\0");
			INT ir = mbstowcs(uzTemp, gzPrincipal, _tcslen(gzPrincipal));
			if (!ir) { hr = HRESULT_FROM_WIN32(GetLastError()); throw hr; }
			ir = SysReAllocString(pbsSamAccountName, uzTemp);
			if (!ir) { hr = HRESULT_FROM_WIN32(GetLastError()); throw hr; }
			uzTemp = _T("\0");  // or
			wcscpy(uzTemp, _T("\0"));
			ir = mbstowcs(uzTemp, gzAuthority, _tcslen(gzAuthority));
			if (!ir) { hr = HRESULT_FROM_WIN32(GetLastError()); throw hr; }
			ir = SysReAllocString(pbsSamAccountName, uzTemp);
			if (!ir) { hr = HRESULT_FROM_WIN32(GetLastError()); throw hr; }
		#endif
	}

	catch(...) {
		OutputDebugStringEx(hr, _bstr_t(_T("GetOwnerPsdk")));
		if (SUCCEEDED(hr)) throw;  // pass unexpected exceptions on to the outer block
	}

	IErrorInfo *pErrorInfo = NULL;
	if (hProcessToken && (INVALID_HANDLE_VALUE != hProcessToken))  {
		BOOL br = CloseHandle(hProcessToken);
		if (!br) throw;
		hProcessToken = NULL;
	}
	if (hProcess && (INVALID_HANDLE_VALUE != hProcess))  {
		BOOL br = CloseHandle(hProcess);
		if (!br) throw;
		hProcess = NULL;
	}

	return hr;
}


HRESULT GetUser(BSTR bsSamAccountName, BSTR bsNetbiosDomain, BSTR *pbsUser)
{
	// purpose = if no domain exists or a domain exists but w/o the use of per OU upnSuffix partitioning then produce
	// a user name with the format of samAccountName@netbiosDomain, optionally could/should be updated to detect if a 
	// domain exists with or w/o per OU upnSuffix partitioning and if so use the samAccountName to call into a cached
	// adodb.connection to the directory to search for the user and extract either their guid and/or upn property as
	// the value that is used for the pbsUser property
	HRESULT hr = S_OK;

	try {
		// validate parameters
		if (!bsSamAccountName || !SysStringLen(bsSamAccountName) || 
			!bsNetbiosDomain || !SysStringLen(bsNetbiosDomain) || !pbsUser) { hr = E_INVALIDARG; throw hr; }
		
		// for now just assume no domain or domain w/o per OU upnSuffix partitioning is present
		_bstr_t sTemp = bsSamAccountName; 
		sTemp += _T("@"); 
		sTemp += bsNetbiosDomain;
		
		INT ir = SysReAllocString(pbsUser, sTemp);
		if (!ir) { hr = HRESULT_FROM_WIN32(GetLastError()); throw hr; }
	}
	
	catch(...) {
		OutputDebugStringEx(hr, _bstr_t(_T("GetUser")));
		if (SUCCEEDED(hr)) throw;  // pass unexpected exceptions on to the outer block
	}

	return hr;
}


HRESULT WriteToLocalLogFile(BSTR bsUrl, BSTR bsXml, BSTR bsSep)
{
	// purpose = write url and xml event payload into a daily local log file for cases where network outage or
	// congestion prevented the submission of an event to the centralized log

	HRESULT hr = S_OK;
	BSTR bsCurrentYmd = NULL;
	try {
		// validate parameters
		if (!bsUrl || !SysStringLen(bsUrl) || !bsXml /*|| !SysStringLen(bsXml)*/) { hr = E_INVALIDARG; throw hr; }
		
		// note - always wrap critical section resource access with local exception handler so
		// we can guarantee that the LeaveCriticalSection call is made no matter how we exit routine

		// first get the log file name that should be currently in use
		SYSTEMTIME oCurrentSystemTime = {0};
		// for log filename use systemtime structure which uses GMT or UTC offset of 0 so that all time
		// values used for filenames will be based on gmt rather than the host specific time zone
		GetSystemTime(&oCurrentSystemTime);  // this returns gmt or utc offset of 0
		if (0 == oCurrentSystemTime.wYear) { hr = HRESULT_FROM_WIN32(GetLastError()); throw hr; }
		// convert systemtime to sql datetime datatype which is date and time data ranging from January 1, 1900,
		// through June 6, 2079, with accuracy to 3.33 milliseconds.  Submitting log event data to sql using native
		// wmi datetime generates errors
		hr = GetYearMonthDay(oCurrentSystemTime, &bsCurrentYmd);
		if (FAILED(hr)) throw hr;

		if (!_sGpecDllPath.length()) {
			// get a path to be used by the WriteToLocalLogFile routine when when xml/http submission of event data fails
			// due to network outages or hiccups, in this case we will just use the same path that the wbem event consumer
			// provider dll is currently registered as being stored under
			_bstr_t sKeyName(_T("CLSID\\"));
			sKeyName += WBEM_EVENT_CSMR_PVDR_CLSID;
			sKeyName += _T("\\InprocServer32");
			CRegKey oRegKey;
			LONG lr = oRegKey.Open(HKEY_CLASSES_ROOT, sKeyName, KEY_ALL_ACCESS);
			if (lr) { hr = HRESULT_FROM_WIN32(lr); throw hr; }  // TODO: comment out throw for vs.net build support
			
			DWORD dwCount = MAX_PATH;  // ERROR_MORE_DATA returned if not large enough
			_TCHAR gzValue[MAX_PATH] = _T("\0");
			lr = oRegKey.QueryValue(gzValue, NULL, &dwCount);  
			// note - when lpszValueName parameter is passed as NULL or an empty string, i.e. _T(""), the method retrieves
			// the type and data for the key's unnamed or default value, if any. This is the value displayed by the 
			// <No Name> entry when viewed by regedt32.exe
			if (lr) { hr = HRESULT_FROM_WIN32(lr); throw hr; }

			_TCHAR *pgzr = _tcsrchr(gzValue, _T('\\'));
			if (!pgzr) { hr = HRESULT_FROM_WIN32(GetLastError()); throw hr; }
			_TCHAR gzTemp[MAX_PATH] = _T("\0");
			#if defined(_DEBUG)
				size_t iLength = (pgzr - gzValue);
			#endif
			_sGpecDllPath = _tcsncpy(gzTemp, gzValue, (pgzr - gzValue));
			if (!_sGpecDllPath.length()) { hr = HRESULT_FROM_WIN32(GetLastError()); throw hr; }

			#if defined(_DEBUG)
				// optional override with hardcoded value to force storage of local log file somewhere else/specific 
				// during debugging work
				// _sGpecDllPath = _T("d:\\vssWorkFld\\ops\\monitoring\\gnlPpsEvntCsmr\\DebugU");
			#endif
		}

		_bstr_t sLogFileName(_sGpecDllPath);
		sLogFileName += _T("\\");
		sLogFileName += LOCAL_LOG_FILE_PREFIX;
		sLogFileName += bsCurrentYmd;
		sLogFileName += _T(".log");

		// grab ownership of log file resource critical section object
		EnterCriticalSection(&_oLogFileCritSctn); 
		hr = HRESULT_FROM_WIN32(GetLastError());
		if (FAILED(hr)) throw hr;

		if (!_spFs.GetInterfacePtr()) {
			// initialize file system object that will be used by WriteToLocalLogFile calls
			// when errors occur attempting to write event records to the centralized rdms
			// hr = _spFs.CreateInstance(__uuidof(IFileSystem), NULL);  // don't use returns hr == 0x80040154 Class not registered
			hr = _spFs.CreateInstance(_T("Scripting.FileSystemObject"), NULL);
			if (FAILED(hr)) throw hr;  // TODO: comment out throw for vs.net build support
		}

		// check if the daily log file for today is currently the one in use and if not then drop 
		// the handle to yesterday's log file and create/attach it to today's daily log file
		VARIANT_BOOL vbr = _spFs->FileExists(sLogFileName);
		if (!vbr) { 
			if (NULL != _spTxtStrm) {
				hr = _spTxtStrm->Close();
				if (FAILED(hr)) throw hr;
				_spTxtStrm.Release();
			}
		}
		
		#if defined(_DEBUGFOO)
			if (NULL != _spTxtStrm) {
				hr = _spTxtStrm->Close();
				if (FAILED(hr)) throw hr;
				_spTxtStrm.Release();
			}
		#endif

		if (NULL == _spTxtStrm) {
			_spTxtStrm = _spFs->OpenTextFile(sLogFileName, ForAppending, VARIANT_BOOL(TRUE), TristateUseDefault);
			if (FAILED(hr)) throw hr;
		}

		// use csv to append xml body so that event record submissions can be easily replayed line by line at a later time
		_bstr_t sLogRecord = bsUrl; 
		if (SysStringLen(bsXml)) { 
			sLogRecord += bsSep; 
			sLogRecord += bsXml;
		}
		hr = _spTxtStrm->WriteLine(sLogRecord);
		if (FAILED(hr)) throw hr;
	}
	
	catch(...) {
		OutputDebugStringEx(hr, _bstr_t(_T("WriteToLocalLogFile")));
		if (SUCCEEDED(hr)) throw;  // pass unexpected exceptions on to the outer block
	}

	// release ownership of the log file resource critical section.
	LeaveCriticalSection(&_oLogFileCritSctn);
	hr = HRESULT_FROM_WIN32(GetLastError());
	if (FAILED(hr)) throw;  // TODO: comment out for vs.net build support

	if (bsCurrentYmd) { SysFreeString(bsCurrentYmd); bsCurrentYmd = NULL; }

	return hr;
}

