// Wuos.cpp : Implementation of CWuos
#include "stdafx.h"
#include "GnlPpsEvntCsmr.h"
#include "Wuos.h"

/////////////////////////////////////////////////////////////////////////////
// CWuos


STDMETHODIMP CWuos::IndicateToConsumer(IWbemClassObject *pLogicalConsumer, LONG lNumObjects, IWbemClassObject **apObjects)
{
	// purpose = process event object and any embedded event objects
	// note - if this method fails or returns a failure code, including GPFs from called routines, CIMOM will recreate
	// the event object and call this method again. If you see this method being called twice for every indication, 
	// it means this method is failing or returning a failure code somehow.  Especially watch the AddRef()/Release()
	// semantics for the embedded object(s).  If they're too low, you'll return a GPF.  The use smartpointers for 
	// interfaces and exception handling simplifies ensuring that these situations do not occur.
	HRESULT hr = S_OK;
	SAFEARRAY *psaEventRecord = NULL;
	BSTR bsTemp = NULL, bsSamAccountName = NULL, bsNetbiosDomain = NULL, bsUser = NULL;
	try {
		// validate parameters
		if (!pLogicalConsumer || !lNumObjects || !apObjects) { hr = E_INVALIDARG; throw hr; }

		// extract some of the logical consumer properties
		_variant_t vTemp;
		hr = pLogicalConsumer->Get(_T("name"), 0L, &vTemp, NULL, NULL);
		if (FAILED(hr)) throw hr;
		_bstr_t sName = vTemp.bstrVal;

		hr = pLogicalConsumer->Get(_T("description"), 0L, &vTemp, NULL, NULL);
		if (FAILED(hr)) throw hr;
		_bstr_t sDescription = vTemp.bstrVal;

		hr = pLogicalConsumer->Get(_T("rdmsHost"), 0L, &vTemp, NULL, NULL);
		if (FAILED(hr)) throw hr;
		_bstr_t sRdmsHost = vTemp.bstrVal;

		hr = pLogicalConsumer->Get(_T("rdmsDb"), 0L, &vTemp, NULL, NULL);
		if (FAILED(hr)) throw hr;
		_bstr_t sRdmsDb = vTemp.bstrVal;

		hr = pLogicalConsumer->Get(_T("rdmsTbl"), 0L, &vTemp, NULL, NULL);
		if (FAILED(hr)) throw hr;
		_bstr_t sRdmsTbl = vTemp.bstrVal;

		hr = pLogicalConsumer->Get(_T("cmdLine"), 0L, &vTemp, NULL, NULL);
		if (FAILED(hr)) throw hr;
		_bstr_t sCmdLine = vTemp.bstrVal;

		WORD wShowWindow = SW_SHOWDEFAULT;
		BOOL bPropsAsArgs = FALSE;
		if (sCmdLine.length()) {
			hr = pLogicalConsumer->Get(_T("cmdLineShwWnd"), 0L, &vTemp, NULL, NULL);
			if (FAILED(hr)) throw hr;
			wShowWindow = vTemp.uiVal;
			hr = pLogicalConsumer->Get(_T("propsAsArgs"), 0L, &vTemp, NULL, NULL);
			if (FAILED(hr)) throw hr;
			bPropsAsArgs = vTemp.boolVal;
		}

		hr = pLogicalConsumer->Get(_T("hostnameProp"), 0L, &vTemp, NULL, NULL);
		if (FAILED(hr)) throw hr;
		_bstr_t sHostnameProp = vTemp.bstrVal;

		hr = pLogicalConsumer->Get(_T("createProp"), 0L, &vTemp, NULL, NULL);
		if (FAILED(hr)) throw hr;
		_bstr_t sCreateProp = vTemp.bstrVal;

		hr = pLogicalConsumer->Get(_T("terminateProp"), 0L, &vTemp, NULL, NULL);
		if (FAILED(hr)) throw hr;
		_bstr_t sTerminateProp = vTemp.bstrVal;

		hr = pLogicalConsumer->Get(_T("otherProps"), 0L, &vTemp, NULL, NULL);
		if (FAILED(hr)) throw hr;
		_bstr_t sOtherProps = vTemp.bstrVal;

		// create safearray of column names and values for passing to IActions:LogEvnt
		// elements = hostname, ipaddress, agent (wmi), name, description, create, terminate, duration, <otherProps>
		// or itg   = node, nodeAlias (ipaddress), manager (snmp | syslog | wmi | ??), alertKey, Summary (could later normalize for storage conservation), datetime (so as to report firstOccurrence, lastOccurrence and tally)
		SAFEARRAYBOUND saBounds[SA_EVENTRECORD_DIMS];
		for (INT iIndex = 0; iIndex < SA_EVENTRECORD_DIMS; iIndex++) {
			saBounds[iIndex].lLbound = SA_EVENTRECORD_LWRBOUND;
			saBounds[iIndex].cElements = SA_EVENTRECORD_ELEMENTS;
		}
		// psaEventRecord = SafeArrayCreate(VT_BSTR, SA_EVENTRECORD_DIMS, saBounds);  // VT_BSTR causes E_OUTMEMORY on SafeArrayPutElement calls
		psaEventRecord = SafeArrayCreate(VT_VARIANT, SA_EVENTRECORD_DIMS, saBounds);
		if (!psaEventRecord) { hr = HRESULT_FROM_WIN32(GetLastError()); throw hr; }
	
		// Iterate through event array of IWbemClassObjects and retrieve desired properties out of the embedded
		// "TargetInstance" object that have been defined by logical consumer otherProps string.  Write these values
		// into a safearray that will be passed to a method that can log the event via xml either directly or indirectly
		// using msmq.

		for (iIndex = 0; iIndex < 1 /* lNumObjects */; iIndex++)  {  // note - usually 1 == lNumObjects
			// define scope outside of if/else conditions
			IWbemClassObjectPtr spTargetInstance;
			
			// note - GpepEvents and likely other __ExtrinsicEvent providers that don't generate events as a result of 
			// instance providers creating a new instance which is passed in events via the TargeteInstance object and
			// subsequently holds all the event properties, instead __ExtrinsicEvent provier events pass their event
			// properties as part of apObject[iIndex] versus apObject[1]->Get(_T("TargetInstance")); so we need to 
			// do a little trick here to work for both __IntrinsicEvent and __ExtrinsicEvent providers

			BOOL bGpepEvent = !_tcscmp(GPEP_EVENT_NAME, sName);
			if (!bGpepEvent) {  // i.e. event with a TargetInstance object containing detailed event properties
				// Get the "TargetInstance" object which defines the specifics about the event that fired.  This will be
				// an embedded object of type VT_UNKNOWN.  All wmi interfaces are derived from IUnknown (aka Custom Interface)
				// versus IDispatch (aka Dual Inteface).  Getting the embedded object "PreviousInstance", if it is present,
				// will enable access to property values that were present before this event fired.  This latter mentioned
				// embedded object is usually present for events of type __InstanceModificationEvent.
				vTemp.Clear();
				hr = apObjects[iIndex]->Get(_T("TargetInstance"), 0L, &vTemp, NULL, NULL);
				if (FAILED(hr)) throw hr;
				
				// Now pull the IUnknown out of the variant. Dont cast directly to IWbemClassObject as it MIGHT work now but
				// a subtle change later will break your code. The PROPER way is to go through QueryInterface and do all
				// the right Release()'s.  Attach the IUnknown from our variant with AddRef = true since we acquired the
				// interface without using smart pointer createinstance() or queryinterface() methods which normally do 
				// the initial AddRef for us
				IUnknownPtr spUnknown;
				// pIUnknown.Attach((IUnknown *)V_UNKNOWN(&vTemp));  // takes ownership of the raw interface pointer
				spUnknown.Attach((IUnknown *)vTemp.punkVal, TRUE);  // AddRef is called
				
				// ask for the IWbemClassObject which is the embedded object.  This will be the instance that was created.
				// IWbemClassObjectPtr spTargetInstance;
				hr = spUnknown.QueryInterface(IID_IWbemClassObject, &spTargetInstance);
				if (FAILED(hr)) throw hr;
				spUnknown.Release();  // since we are done with IUnknown we will release it
			}
			else  {  // i.e. event without a TargetInstance object containing detailed properties
				// assign event object to IWbemClassObjectPtr spTargetInstance so that remainder of routine used to
				// extract event properties works the same for event with properties buried in a TargetInstance object
				// as well as those events with properties held directly on event object itself

				// IWbemClassObjectPtr spTargetInstance;
				// spTargetInstance.Attach(pLogicalConsumer, TRUE);  // AddRef is called
				spTargetInstance.Attach(apObjects[iIndex], TRUE);  // AddRef is called
				if (!spTargetInstance.GetInterfacePtr()) { hr = E_INVALIDARG; throw hr; } // ???
			}

			// acquire and write the column 0 hostname value into safearray
			BOOL bHostnamePropValueProvided = FALSE;
			if (_tcscmp(_T(""), sHostnameProp)) {
				vTemp.Clear(); 
				hr = spTargetInstance->Get(sHostnameProp, 0L, &vTemp, NULL, NULL);
				if (FAILED(hr)) throw hr;
				// now do something to address IGpep::Fire optional vHostname parameter cases
				if ((VT_EMPTY == vTemp.vt) || (VT_NULL == vTemp.vt) || !_tcscmp(_T(""), vTemp.bstrVal))
					bHostnamePropValueProvided = FALSE;
				else
					bHostnamePropValueProvided = TRUE;
			}
			// else { 
			if (!bHostnamePropValueProvided) {
				// winsock method for acquiring dns host name, does not support generic text
				/* 
				// note - see DllMain for winsock data structure startup call
				WSADATA oWsaData = {0};
				INT ir = WSAStartup(MAKEWORD(1, 1), &oWsaData);
				if (ir) { hr = HRESULT_FROM_WIN32(WSAGetLastError()); throw hr; }
				*/
				char szTemp[MAX_HOSTNAME] = "\0";
				INT ccTemp = MAX_HOSTNAME;
				INT ir = gethostname(szTemp, ccTemp);
				if (ir) { hr = HRESULT_FROM_WIN32(WSAGetLastError()); throw hr; }
				vTemp.Clear(); vTemp = _bstr_t(szTemp);
				/*
				// note - see DllMain for winsock data structure cleanup call
				ir = WSACleanup();
				if (ir) { hr = HRESULT_FROM_WIN32(WSAGetLastError()); throw hr; }
				*/
				
				/*
				// psdk method for acquring dns hostname, supports generic text but isn't recognized by compiler ???
				_TCHAR gzTemp[MAX_HOSTNAME] = _T("\0");
				#if defined(_DEBUG)
					ccTemp = MAX_HOSTNAME;
				#else
					INT ccTemp = MAX_HOSTNAME;
				#endif
				BOOL br = GetComputerNameEx(ComputerNameDnsDomaineType, // name type
											gzTemp,  // name buffer
											&ccTemp);  // count of characters in buffer
				if (!br) { hr = HRESULT_FROM_WIN32(GetLastError()); throw hr; }
				vTemp.Clear(); vTemp = _bstr_t(gzTemp);
				*/
			}
			INT iX = 0;  // column index
			LONG alCoords[2] = {iX, 0};  // [0] == x coord;  [1] == y coord
			hr = SafeArrayPutElement(psaEventRecord, alCoords, &_variant_t(COLUMN_0_NAME));
			if (FAILED(hr)) throw hr;
			alCoords[0] = 0;  alCoords[1] = 1;  // [0] == x coord;  [1] == y coord
			hr = SafeArrayPutElement(psaEventRecord, alCoords, &vTemp);
			if (FAILED(hr)) throw hr;
			if (bPropsAsArgs) { sCmdLine += _T(" "); sCmdLine += vTemp.bstrVal; }

			// acquire and write the column 1 ipAddress value into safearray
			// winsock method for acquiring ip address for a given hostname, does not support generic text
			/* note - see header file ObjectMain for winsock data structure startup call
			WSADATA oWsaData = {0};
			INT ir = WSAStartup(MAKEWORD(1, 1), &oWsaData);
			if (ir) { hr = HRESULT_FROM_WIN32(WSAGetLastError()); throw hr; }
			*/
			HOSTENT *pHostEnt = NULL;
			char szHostname[MAX_HOSTNAME] = "\0";
			_TCHAR gzIpAddress[MAX_IPADDRESS] = _T("\0");
			INT ccHostname = MAX_HOSTNAME;
			size_t str = wcstombs(szHostname, (WCHAR *)vTemp.bstrVal, ccHostname);
			// if (-1 == str) {hr = <some wcs characters could not be converted to mbs)
			pHostEnt = gethostbyname(szHostname);
			if (pHostEnt) {
				int iAdapter = 0, iIpIndex = 0;
				BOOL bIpAddressFound = FALSE;
				while (pHostEnt->h_addr_list[iAdapter] && !bIpAddressFound) {  // loop causes last Adapter/ipAddress to be the one that is used
					if(127 != (BYTE)pHostEnt->h_addr_list[iAdapter][0]) {
						// extract the current address in host order
						_stprintf(gzIpAddress, _T("%d.%d.%d.%d"), (BYTE)pHostEnt->h_addr_list[iAdapter][0], (BYTE)pHostEnt->h_addr_list[iAdapter][1], (BYTE)pHostEnt->h_addr_list[iAdapter][2], (BYTE)pHostEnt->h_addr_list[iAdapter][3]);
						bIpAddressFound = TRUE; // just pick off the first non-loopback address since it will be the primary
					}
					iAdapter++;
				}
			}
			else {  // don't fail just default to an all zero's ip address
				_tcscpy(gzIpAddress, _T("0.0.0.0\0"));
			}
			/* note - see header file ObjectMain for winsock data structure cleanup call
			ir = WSACleanup();
			if (ir) { hr = HRESULT_FROM_WIN32(WSAGetLastError()); throw hr; }
			*/
			vTemp.Clear();	vTemp = _bstr_t(gzIpAddress);
			iX++;  // column index
			alCoords[0] = iX;  alCoords[1] = 0;  // [0] == x coord;  [1] == y coord
			hr = SafeArrayPutElement(psaEventRecord, alCoords, &_variant_t(COLUMN_1_NAME));
			if (FAILED(hr)) throw hr;
			alCoords[0] = iX;  alCoords[1] = 1;  // [0] == x coord;  [1] == y coord
			hr = SafeArrayPutElement(psaEventRecord, alCoords, &vTemp);
			if (FAILED(hr)) throw hr;
			if (bPropsAsArgs) { sCmdLine += _T(" "); sCmdLine += vTemp.bstrVal; }

			// acquire and write the column 2 agent value into safearray
			vTemp.Clear();	vTemp = _bstr_t(GPEC_VERSION);
			iX++;  // column index
			alCoords[0] = iX;  alCoords[1] = 0;  // [0] == x coord;  [1] == y coord
			hr = SafeArrayPutElement(psaEventRecord, alCoords, &_variant_t(COLUMN_2_NAME));
			if (FAILED(hr)) throw hr;
			alCoords[0] = iX;  alCoords[1] = 1;  // [0] == x coord;  [1] == y coord
			hr = SafeArrayPutElement(psaEventRecord, alCoords, &vTemp);
			if (FAILED(hr)) throw hr;
			if (bPropsAsArgs) { sCmdLine += _T(" "); sCmdLine += vTemp.bstrVal; }

			// acquire and write the column 3 logical consumer name value into safearray
			vTemp.Clear();	vTemp = sName;
			iX++;  // column index
			alCoords[0] = iX;  alCoords[1] = 0;  // [0] == x coord;  [1] == y coord
			hr = SafeArrayPutElement(psaEventRecord, alCoords, &_variant_t(COLUMN_3_NAME));
			if (FAILED(hr)) throw hr;
			alCoords[0] = iX;  alCoords[1] = 1;  // [0] == x coord;  [1] == y coord
			hr = SafeArrayPutElement(psaEventRecord, alCoords, &vTemp);
			if (FAILED(hr)) throw hr;
			if (bPropsAsArgs) { sCmdLine += _T(" "); sCmdLine += vTemp.bstrVal; }

			// acquire and write the column 4 logical consumer description value into safearray
			vTemp.Clear(); vTemp = sDescription;
			iX++;  // column index
			alCoords[0] = iX;  alCoords[1] = 0;  // [0] == x coord;  [1] == y coord
			hr = SafeArrayPutElement(psaEventRecord, alCoords, &_variant_t(COLUMN_4_NAME));
			if (FAILED(hr)) throw hr;
			alCoords[0] = iX;  alCoords[1] = 1;  // [0] == x coord;  [1] == y coord
			hr = SafeArrayPutElement(psaEventRecord, alCoords, &vTemp);
			if (FAILED(hr)) throw hr;
			if (bPropsAsArgs) { sCmdLine += _T(" "); sCmdLine += vTemp.bstrVal; }

			// acquire and write the column 5 create value into safearray
			BOOL bCreatePropValueProvided = FALSE;
			SYSTEMTIME oCreateSystemTime = {0};
			if (_tcscmp(_T(""), sCreateProp)) {
				// wmi datetime format = yyyymmddHHMMSS.mmmmmmsUUU
				// mmmmmm = six digit representation of microseconds
				// s = + | - offset from UTC (GMT has UTC offset of 0)
				// UUU = three digit representation of the number of minutes that originating time zone deviates from UTC
				// for WMI, it is encouraged, but not required, to convert times to GMT with a UTC offset of zero.
				vTemp.Clear();
				hr = spTargetInstance->Get(sCreateProp, 0L, &vTemp, NULL, NULL);
				if (FAILED(hr)) throw hr;
				// now do something to address IGpep::Fire optional vCreationDate parameter cases
				#if defined(_DEBUG)
					/*
					_bstr_t sr(DEFAULT_WMIDATETIME);
					INT ir = _tcscmp(sr, vTemp.bstrVal);
					ir = !_tcscmp(DEFAULT_WMIDATETIME, vTemp.bstrVal);
					*/
					INT ir = _tcscmp(DEFAULT_WMIDATETIME, vTemp.bstrVal);
					ir = !_tcscmp(DEFAULT_WMIDATETIME, vTemp.bstrVal);
				#endif
				if ((VT_EMPTY == vTemp.vt) || (VT_NULL == vTemp.vt) ||
					(!_tcscmp(_T(""), vTemp.bstrVal)) || (!_tcscmp(DEFAULT_WMIDATETIME, vTemp.bstrVal)))
					bCreatePropValueProvided = FALSE;
				else {
					bCreatePropValueProvided = TRUE;
					// convert wmi datetime to systemtime structure which uses GMT or UTC offset of 0 so that all time
					// values stored in database will be based on gmt of event rather that host specific time zone
					hr = GetSystemTimeEx(vTemp.bstrVal, &oCreateSystemTime);
					if (FAILED(hr)) throw hr;
				}
			}
			// else {
			if (!bCreatePropValueProvided) {
				GetSystemTime(&oCreateSystemTime);  // this returns gmt or utc offset of 0
				if (0 == oCreateSystemTime.wYear) { hr = HRESULT_FROM_WIN32(GetLastError()); throw hr; }
			}
			
			// convert systemtime to sql datetime datatype which is date and time data ranging from January 1, 1900,
			// through June 6, 2079, with accuracy to 3.33 milliseconds.  Submitting log event data to sql using native
			// wmi datetime generates errors
			hr = GetSqlDatetime(oCreateSystemTime, &bsTemp);
			if (FAILED(hr)) throw hr;

			vTemp.Clear();	vTemp = bsTemp; SysFreeString(bsTemp); bsTemp = NULL;
			iX++;  // column index
			alCoords[0] = iX;  alCoords[1] = 0;  // [0] == x coord;  [1] == y coord
			hr = SafeArrayPutElement(psaEventRecord, alCoords, &_variant_t(COLUMN_5_NAME));
			if (FAILED(hr)) throw hr;
			alCoords[0] = iX;  alCoords[1] = 1;  // [0] == x coord;  [1] == y coord
			hr = SafeArrayPutElement(psaEventRecord, alCoords, &vTemp);
			if (FAILED(hr)) throw hr;
			if (bPropsAsArgs) { sCmdLine += _T(" "); sCmdLine += vTemp.bstrVal; }

			// acquire and write the column 6 terminate value into safearray
			// note - column 6 dtTerminate is applicable for gpepEvent events where TerminateProp is defined to be
			// "TerminationDate" and the value of this property for the current event object is not NULL.  Column 6
			// dtTerminate is also applicable for prcTerminate events where TerminateProp is never defined because the
			// w2ks rtm wmi Win32_Process instance provider currently does not populate "TerminationDate" property of 
			// the TargetInstance object so we always calculate the TerminationDate using a GetSystemTime call
			BOOL bGpepTerminateEvent = FALSE;
			// _tcscmp(_T(""), sTerminateProp) ? bGpepTerminateEvent = TRUE : bGpepTerminateEvent = FALSE;
			sTerminateProp.length() ? bGpepTerminateEvent = TRUE : bGpepTerminateEvent = FALSE;

			BOOL bPrcTerminateEvent = FALSE;
			(!bGpepTerminateEvent && !_tcsicmp(PRC_TERMINATE_EVENT_DESCRIPTION, sDescription)) ? bPrcTerminateEvent = TRUE : bPrcTerminateEvent = FALSE;

			// define scope outside of two if conditionals that use variable
			SYSTEMTIME oTerminateSystemTime = {0};
			if (bGpepTerminateEvent) {
				// wmi datetime format = yyyymmddHHMMSS.mmmmmmsUUU
				// mmmmmm = six digit representation of microseconds
				// s = + | - offset from UTC (GMT has UTC offset of 0)
				// UUU = three digit representation of the number of minutes that originating time zone deviates from UTC
				// for WMI, it is encouraged, but not required, to convert times to GMT with a UTC offset of zero.
				vTemp.Clear(); vTemp = _T("");
				hr = spTargetInstance->Get(sTerminateProp, 0L, &vTemp, NULL, NULL);
				if (FAILED(hr)) throw hr;
				// now do something to address IGpep::Fire optional vTerminationDate parameter cases
				#if defined(_DEBUG)
					/*
					_bstr_t sr(DEFAULT_WMIDATETIME);
					INT ir = _tcscmp(sr, vTemp.bstrVal);
					ir = !_tcscmp(DEFAULT_WMIDATETIME, vTemp.bstrVal);
					*/
					INT ir = _tcscmp(DEFAULT_WMIDATETIME, vTemp.bstrVal);
					ir = !_tcscmp(DEFAULT_WMIDATETIME, vTemp.bstrVal);
				#endif
				if ((VT_EMPTY == vTemp.vt) || (VT_NULL == vTemp.vt) || 
					(!_tcscmp(_T(""), vTemp.bstrVal)) || (!_tcscmp(DEFAULT_WMIDATETIME, vTemp.bstrVal))) {
					 // sTerminateProp was provided but no value was set, i.e. GpepEvent was a generic or createEvent
					bGpepTerminateEvent = FALSE;  
				}
				else { 
					bGpepTerminateEvent = TRUE;  
					// convert wmi datetime to systemtime structure which uses GMT or UTC offset of 0 so that all time
					// values stored in database will be based on gmt of event rather that host specific time zone
					hr = GetSystemTimeEx(vTemp.bstrVal, &oTerminateSystemTime);
					if (FAILED(hr)) throw hr;
				}
			}

			if (bPrcTerminateEvent) {
				GetSystemTime(&oTerminateSystemTime);  // this returns gmt or utc offset of 0
				if (0 == oTerminateSystemTime.wYear) { hr = HRESULT_FROM_WIN32(GetLastError()); throw hr; }
			}

			if (bGpepTerminateEvent || bPrcTerminateEvent) {
				// convert systemtime to sql datetime datatype which is date and time data ranging from January 1, 1900,
				// through June 6, 2079, with accuracy to 3.33 milliseconds.  Submitting log event data to sql using native
				// wmi datetime generates errors
				hr = GetSqlDatetime(oTerminateSystemTime, &bsTemp);
				if (FAILED(hr)) throw hr;

				vTemp.Clear(); vTemp = bsTemp; SysFreeString(bsTemp); bsTemp = NULL;
				iX++;  // column index
				alCoords[0] = iX;  alCoords[1] = 0;  // [0] == x coord;  [1] == y coord
				hr = SafeArrayPutElement(psaEventRecord, alCoords, &_variant_t(COLUMN_6_NAME));
				if (FAILED(hr)) throw hr;
				alCoords[0] = iX;  alCoords[1] = 1;  // [0] == x coord;  [1] == y coord
				vTemp.ChangeType(VT_BSTR, NULL);
				hr = SafeArrayPutElement(psaEventRecord, alCoords, &vTemp);
				if (FAILED(hr)) throw hr;
				if (bPropsAsArgs) { sCmdLine += _T(" "); sCmdLine += vTemp.bstrVal; }

				// acquire and write, for denormalization/reporting purposes, the column 7 duration value into safearray
				LONG lDuration = 0L;
				hr = GetSystemTimeElapsedInSeconds(oCreateSystemTime, oTerminateSystemTime, &lDuration);
				if (FAILED(hr)) throw hr;

				vTemp.Clear();	vTemp = lDuration; vTemp.ChangeType(VT_BSTR, NULL);
				iX++;  // column index
				alCoords[0] = iX;  alCoords[1] = 0;  // [0] == x coord;  [1] == y coord
				hr = SafeArrayPutElement(psaEventRecord, alCoords, &_variant_t(COLUMN_7_NAME));
				if (FAILED(hr)) throw hr;
				alCoords[0] = iX;  alCoords[1] = 1;  // [0] == x coord;  [1] == y coord
				hr = SafeArrayPutElement(psaEventRecord, alCoords, &vTemp);
				if (FAILED(hr)) throw hr;
				if (bPropsAsArgs) { sCmdLine += _T(" "); sCmdLine += vTemp.bstrVal; }
			}

			// acquire and write the column 8 user value into safearray
			// note - column 8 szUser is applicable for gpepEvent events where the GPEP_EVENT_USER_PROP contains the
			// applicable user value directly.  Column 8 szUser is also applicable for prcCreate events where the
			// applicable user value must be extracted by first acquiring the WIN32_PROCESS_PROCESSID_PROP value and
			// using this ProcessID value to call into the win32 api's and extract the applicable user value.
			if (!_tcsicmp(GPEP_EVENT_DESCRIPTION, sDescription)) {
				// get the owner of the GpepEvent instance
				vTemp.Clear(); 
				hr = spTargetInstance->Get(GPEP_EVENT_USER_PROP, 0L, &vTemp, NULL, NULL);
				if (FAILED(hr)) throw hr;
				vTemp.ChangeType(VT_BSTR, NULL);  // make sure it is in bstr format

				iX++;  // column index
				alCoords[0] = iX;  alCoords[1] = 0;  // [0] == x coord;  [1] == y coord
				hr = SafeArrayPutElement(psaEventRecord, alCoords, &_variant_t(COLUMN_8_NAME));
				if (FAILED(hr)) throw hr;
				alCoords[0] = iX;  alCoords[1] = 1;  // [0] == x coord;  [1] == y coord
				hr = SafeArrayPutElement(psaEventRecord, alCoords, &vTemp);
				if (FAILED(hr)) throw hr;
				if (bPropsAsArgs) { sCmdLine += _T(" "); sCmdLine += vTemp.bstrVal; }
			}

			if (!_tcsicmp(PRC_CREATE_EVENT_DESCRIPTION, sDescription)) {	
				/*
				// get the owner of the Win32_Process instance, the wmi win32_process method provider way which
				// word has it doesn't work because the caller of the method must be the same as the owner of
				// the Win32_Process instance event against which this method is being called.  Since our consumer
				// runs in the security context of the wmi service this will rarely/never be the case.
				vTemp.Clear();
				hr = spTargetInstance->Get(_T("__Namespace"), 0L, &vTemp, NULL, NULL);
				if (FAILED(hr)) throw hr;
				_bstr_t sNamespace = vTemp.bstrVal;
				
				vTemp.Clear();
				hr = spTargetInstance->Get(_T("__Class"), 0L, &vTemp, NULL, NULL);
				if (FAILED(hr)) throw hr;
				_bstr_t sClassPath = vTemp.bstrVal;
				
				vTemp.Clear();
				// hr = spTargetInstance->Get(_T("__Path"), 0L, &vTemp, NULL, NULL);
				hr = spTargetInstance->Get(_T("__Relpath"), 0L, &vTemp, NULL, NULL);
				if (FAILED(hr)) throw hr;
				_bstr_t sInstancePath = vTemp.bstrVal;

				if (bsSamAccountName) { SysFreeString(bsSamAccountName); bsSamAccountName = NULL; }
				if (bsNetbiosDomain) { SysFreeString(bsNetbiosDomain); bsNetbiosDomain = NULL; }
				hr = GetOwnerWmi(sNamespace, sClassPath, sInstancePath, &bsSamAccountName, &bsNetbiosDomain);
				if (FAILED(hr)) throw hr;
				*/

				// get the owner of the Win32_Process instance, the win32 way
				vTemp.Clear(); 
				hr = spTargetInstance->Get(WIN32_PROCESS_PROCESSID_PROP, 0L, &vTemp, NULL, NULL);
				if (FAILED(hr)) throw hr;
				vTemp.ChangeType(VT_UI4, NULL);  // make sure it is in dword format
				DWORD dwProcessId = vTemp.ulVal;
				if (bsSamAccountName) { SysFreeString(bsSamAccountName); bsSamAccountName = NULL; }
				if (bsNetbiosDomain) { SysFreeString(bsNetbiosDomain); bsNetbiosDomain = NULL; }
				hr = GetOwnerWin32(dwProcessId, &bsSamAccountName, &bsNetbiosDomain);
				if (FAILED(hr)) throw hr;

				// generate a user name to log with record currently using format of samAccountName@netBiosDomain
				// which should enable an efficient compromise for user associated event tracking in cases of 
				// standalone, single domain and multipartitioned domain environments, see GetUse for more details
				if (bsUser) { SysFreeString(bsUser); bsUser = NULL; }
				hr = GetUser(bsSamAccountName, bsNetbiosDomain, &bsUser);
				if (FAILED(hr)) throw hr;
				SysFreeString(bsSamAccountName); bsSamAccountName = NULL; SysFreeString(bsNetbiosDomain); bsNetbiosDomain = NULL;
				
				vTemp.Clear(); vTemp = bsUser; SysFreeString(bsUser); bsUser = NULL;
				iX++;  // column index
				alCoords[0] = iX;  alCoords[1] = 0;  // [0] == x coord;  [1] == y coord
				hr = SafeArrayPutElement(psaEventRecord, alCoords, &_variant_t(COLUMN_8_NAME));
				if (FAILED(hr)) throw hr;
				alCoords[0] = iX;  alCoords[1] = 1;  // [0] == x coord;  [1] == y coord
				hr = SafeArrayPutElement(psaEventRecord, alCoords, &vTemp);
				if (FAILED(hr)) throw hr;
				if (bPropsAsArgs) { sCmdLine += _T(" "); sCmdLine += vTemp.bstrVal; }
			}

			// acquire and write the column iX -> SA_EVENTRECORD_ELEMENTS other property values into safearray
			_TCHAR gzTemp[MAX_OTHERPROPS] = _T("\0");  // copy sOtherProps into buffer because _tcsTok messes with null terminators
			#if defined(_UNICODE)
				LPTSTR pgzr = _tcscpy(gzTemp, sOtherProps);
				if (!pgzr) { hr = HRESULT_FROM_WIN32(GetLastError()); throw hr; }
			#else
				size_t str = wcstombs(gzTemp, sOtherProps, sOtherProps.length());
				if (!str) { hr = HRESULT_FROM_WIN32(GetLastError()); throw hr; }
			#endif
			_TCHAR gzDelimit[] = _T(" ,\n");  // i.e. space, comma or crlf
			LPTSTR pgzToken = NULL;
			pgzToken = _tcstok(gzTemp, gzDelimit);
			iX++;  // column index
			INT iColumnLetterOffset = iX;
			while (pgzToken && (SA_EVENTRECORD_ELEMENTS > iX )) {
				vTemp.Clear();
				hr = spTargetInstance->Get(pgzToken, 0L, &vTemp, NULL, NULL);
				if (FAILED(hr)) throw hr;
				vTemp.ChangeType(VT_BSTR, NULL);
				
				// generate column name using basename and then appending A, B, C, etc. incrementally
				_bstr_t sTemp = COLUMN_OTHERPROPS_BASENAME;
				_TCHAR gzColumnNameSuffix[] = _T("\0");
				gzColumnNameSuffix[0] = ('A' - iColumnLetterOffset + iX);  // 'A' = 64 | 0x41
				gzColumnNameSuffix[1] = '\0';
				sTemp += gzColumnNameSuffix;

				alCoords[0] = iX;  alCoords[1] = 0;  // [0] == x coord;  [1] == y coord
				hr = SafeArrayPutElement(psaEventRecord, alCoords, &_variant_t(sTemp));
				if (FAILED(hr)) throw hr;
				alCoords[0] = iX;  alCoords[1] = 1;  // [0] == x coord;  [1] == y coord
				hr = SafeArrayPutElement(psaEventRecord, alCoords, &vTemp);
				if (FAILED(hr)) throw hr;
				if (bPropsAsArgs) { sCmdLine += _T("\0"); sCmdLine += vTemp.bstrVal; }

				// move to next token and column number in preparation for next iteration of loop
				pgzToken = _tcstok(NULL, gzDelimit);
				iX++;
			}

			INT cfEventRecord = iX;

			// now call LogEvnt consumer action behavior
			vTemp.Clear();
			vTemp.parray = psaEventRecord;  // type should now be VT_ARRAY

			IActionsPtr pActions;
			hr = pActions.CreateInstance(CLSID_Actions, NULL);
			if (FAILED(hr)) throw hr;
			
			if (!bGpepTerminateEvent && !bPrcTerminateEvent) // non-terminateEvents are processed as inserts
				hr = pActions->XmlHttpSqlInsert(sRdmsHost, sRdmsDb, sRdmsTbl, vTemp, cfEventRecord, USE_HTTP_POST);
			else  // terminateEvents are processed as updates
				hr = pActions->XmlHttpSqlUpdate(sRdmsHost, sRdmsDb, sRdmsTbl, vTemp, cfEventRecord, USE_HTTP_POST);
			if (FAILED(hr)) throw hr;
			
			if (sCmdLine.length()) {
				IActionsPtr pActions;
				hr = pActions.CreateInstance(CLSID_Actions, NULL);
				if (FAILED(hr)) throw hr;

				hr = pActions->CreatePrc(sCmdLine, wShowWindow);
				if (FAILED(hr)) throw hr;
				pActions.Release();
			}

			// completed submitting event to centralized log via [msmq/]xml/http
			spTargetInstance.Release();  // done with IWbemClassObject
		
		}  // completed walking embedded object list
			
		// psdk document that your return WBEM_S_NO_ERROR if IndicateToConsumer complete successfully but 
		// since S_OK == WBEM_S_NO_ERROR then doing nothing more work as well
		hr = WBEM_S_NO_ERROR;
	}

	catch(...) {
		#if defined(_DEBUG)
			HRESULT hrTest = HRESULT_FROM_WIN32(GetLastError());
		#endif
		OutputDebugStringEx(hr, _bstr_t(_T("IndicateToConsumer")));
		if (SUCCEEDED(hr)) throw;  // pass unexpected exceptions on to the outer block
	}

	if (bsUser) { SysFreeString(bsUser); bsUser = NULL; }
	if (bsNetbiosDomain) { SysFreeString(bsNetbiosDomain); bsNetbiosDomain = NULL; }
	if (bsSamAccountName) {	SysFreeString(bsSamAccountName); bsSamAccountName = NULL; }
	if (bsTemp) { SysFreeString(bsTemp); bsTemp = NULL; }
	if (psaEventRecord) {
		IErrorInfo *pErrorInfo = NULL;
		HRESULT hr = SafeArrayDestroy(psaEventRecord);
		if (FAILED(hr)) throw;
		psaEventRecord = NULL;
	}

	return hr;
}
